/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/

#if RF_OPT_USE_MALLOC == 0
static rf_volume_comm_t      g_volume[RF_COMM_MAX_VOLUME];
static rf_file_comm_t        g_file[RF_COMM_MAX_FILE];
#else
static list_t* g_volume_list;
static list_t* g_file_list;
#endif

#ifdef RF_DEBUG
static rf_bool_t vol_handle_check(rf_handle_t fh)
{
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;
    for (i = 0; i < RF_COMM_MAX_VOLUME; i++) {
        if (&g_volume[i] == (rf_volume_comm_ref_t)fh)
            return TRUE;
    }
#else
    list_t* entry;
    rf_volume_comm_ref_t vol;

    entry = g_volume_list;
    for (;;) {
        if (entry == NULL)
            break;
        vol = RF_CONTAINING_RECORD(entry, rf_volume_comm_t, list);
        if (vol == (rf_volume_comm_ref_t)fh)
            return TRUE;
        entry = list_item_next(entry);
    }
#endif
    return FALSE;
}
#endif

rf_volume_comm_ref_t vol_alloc(rf_void_t)
{
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;
    for (i = 0; i < RF_COMM_MAX_VOLUME; i++) {
        if (!g_volume[i].open_flag) {
            memset(&g_volume[i], 0, sizeof(g_volume[i]));
            g_volume[i].fs_type = RF_COMM_INIT_FS_TYPE;
            return &g_volume[i];
        }
    }
    return NULL;
#else
    rf_volume_comm_ref_t vol = (rf_volume_comm_ref_t)malloc(sizeof(rf_volume_comm_t));
    if (vol == NULL)
        return NULL;
    memset(vol, 0, sizeof(*vol));
    list_add(&g_volume_list, &vol->list);
    return vol;
#endif
}

rf_void_t vol_free(rf_volume_comm_ref_t vol)
{
    RF_ASSERTE("error: invalid vol handle\n", vol_handle_check(vol));

#if RF_OPT_USE_MALLOC == 0
    vol->open_flag = FALSE;
#else
    list_remove(&g_volume_list, &vol->list);
    rf_free(vol);
#endif
}

rf_void_t init_data(rf_void_t)
{
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;
    for (i = 0; i < RF_COMM_MAX_VOLUME; i++)
        g_volume[i].open_flag = FALSE;

    for (i = 0; i < RF_COMM_MAX_FILE; i++)
        g_file[i].open_flag = FALSE;
#else
    list_init(&g_file_list);
    list_init(&g_volume_list);
#endif
}

#if RF_OPT_USE_GET_HANDLE_COUNT
rf_int_t comm_get_volume_handle_count(rf_int_t* available)
{
    rf_int_t use_cout = 0, available_cout = -1;

#if RF_OPT_USE_MALLOC == 0
    {
        rf_int_t i;
        for (i = 0; i < RF_COMM_MAX_VOLUME; i++) {
            if (g_volume[i].open_flag)
                use_cout++;
        }
    }
    available_cout = RF_COMM_MAX_FILE - use_cout;
#else
    {
        list_t* entry = g_volume_list;
        for (;;) {
            if (entry == NULL)
                break;
            use_cout++;
            entry = list_item_next(entry);
        }
    }
#endif
    if (available)
        *available = available_cout;
    return use_cout;
}

rf_int_t comm_get_file_handle_count(rf_int_t* available)
{
    rf_int_t use_cout = 0, available_cout = -1;

#if RF_OPT_USE_MALLOC == 0
    {
        rf_int_t i;

        for (i = 0; i < RF_COMM_MAX_FILE; i++) {
            if (g_file[i].open_flag)
                use_cout++;
        }
    }
    available_cout = RF_COMM_MAX_FILE - use_cout;
#else
    {
        list_t* entry = g_file_list;
        for (;;) {
            if (entry == NULL)
                break;
            use_cout++;
            entry = list_item_next(entry);
        }
    }
#endif
    if (available)
        *available = available_cout;
    return use_cout;
}
#endif

/* ļȨ Ȩ޿Թ дȨΪռ */
/* ntfs ext2 ֻļϵͳ ͬʱļ */
#ifdef _RF_FAT_MODULE
#if RF_OPT_FILE_ACCESS_CHECK
rf_bool_t comm_check_file_access(rf_file_comm_ref_t file)
{
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;

    for (i = 0; i < RF_COMM_MAX_FILE; i++) {
        if (g_file[i].open_flag && (&g_file[i] != file)) {
            if (file->vol == g_file[i].vol &&
                    file->entry.sector == g_file[i].entry.sector &&
                    file->entry.offset == g_file[i].entry.offset) {
                if (file->access_flag) {
                    /* д ܴһ д */
                    return FALSE;
                } else {
                    /*  ܴһд */
                    if (g_file[i].access_flag & 0x3)
                        return FALSE;
                }
            }
        }
    }
#else
    list_t* entry;
    rf_file_comm_ref_t _file;

    entry = g_file_list;
    for (;;) {
        if (entry == NULL)
            break;
        _file = RF_CONTAINING_RECORD(entry, rf_file_comm_t, list);
        if (_file->open_flag && _file != file) {
            if (file->vol == _file->vol &&
                    file->entry.sector == _file->entry.sector &&
                    file->entry.offset == _file->entry.offset) {
                if (file->access_flag) {
                    /* д ܴһ д */
                    return FALSE;
                } else {
                    /*  ܴһд */
                    if (file->access_flag & 0x3)
                        return FALSE;
                }
            }
        }
        entry = list_item_next(entry);
    }
#endif
    return TRUE;
}
#endif

#if RF_OPT_USE_SET_NAME == 1|| RF_OPT_USE_GET_SET_LONG_NAME == 1
#if RF_OPT_FILE_ACCESS_CHECK
/* ͨĿ¼ļʼ Ȩ޲ */
rf_bool_t check_start_cluster_access(rf_volume_comm_ref_t vol, rf_u32_t clst, rf_int_t flag)
{
    /* Ŀ¼ м */
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;

    for (i = 0; i < RF_COMM_MAX_FILE; i++) {
        if (g_file[i].open_flag) {
            if (vol == g_file[i].vol && clst == g_file[i].start_cluster && g_file[i].start_cluster != 0) {
                if (flag) {
                    /* д ܴһ д */
                    return FALSE;
                } else {
                    /*  ܴһд */
                    if (g_file[i].access_flag)
                        return FALSE;
                }
            }
        }
    }
#else
    list_t* entry;
    rf_file_comm_ref_t file;

    entry = g_file_list;
    for (;;) {
        if (entry == NULL)
            break;
        file = RF_CONTAINING_RECORD(entry, rf_file_comm_t, list);
        if (file->open_flag) {
            if (vol == file->vol && clst == file->start_cluster && file->start_cluster != 0) {
                if (flag) {
                    /* д ܴһ д */
                    return FALSE;
                } else {
                    /*  ܴһд */
                    if (file->access_flag)
                        return FALSE;
                }
            }
        }
        entry = list_item_next(entry);
    }
#endif
    return TRUE;
}
#endif
#endif
#endif

static rf_file_comm_ref_t file_alloc(rf_volume_comm_ref_t vol)
{
    rf_file_comm_ref_t file = NULL;

#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;
    for (i = 0; i < RF_COMM_MAX_FILE; i++) {
        if (g_file[i].open_flag == 0) {
            memset(&g_file[i], 0, sizeof(g_file[i]));
            file = &g_file[i];
            break;
        }
    }
#else
    file = (rf_file_comm_ref_t)malloc(sizeof(rf_file_comm_t));
    if (file != NULL) {
        memset(file, 0, sizeof(*file));
        list_add(&g_file_list, &file->list);
    }
#endif
    if (file == NULL)
        return NULL;
    file->vol = vol;
    file->open_flag = TRUE;
    return file;
}

#ifdef RF_DEBUG
static rf_bool_t file_handle_check(rf_handle_t fh)
{
#if RF_OPT_USE_MALLOC == 0
    rf_int_t i;
    for (i = 0; i < RF_COMM_MAX_FILE; i++) {
        if (&g_file[i] == (rf_file_comm_ref_t)fh)
            return TRUE;
    }
#else
    rf_file_comm_ref_t file;
    list_t* entry = g_file_list;

    for (;;) {
        if (entry == NULL)
            break;
        file = RF_CONTAINING_RECORD(entry, rf_file_comm_t, list);
        if (file == (rf_file_comm_ref_t)fh)
            return TRUE;
        entry = list_item_next(entry);
    }
#endif
    return FALSE;
}
#endif

static rf_void_t file_free(rf_file_comm_ref_t file)
{
#ifdef RF_DEBUG
    rf_bool_t ret;

    RF_ASSERTE("file==NULL", file != NULL);
    ret = file_handle_check(file);
    RF_ASSERTE("file_handle_check err", ret == TRUE);
#endif

#if RF_OPT_USE_MALLOC == 0
    file->open_flag = FALSE;
#else
    list_remove(&g_file_list, &file->list);
    rf_free(file);
#endif
}
