/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "ext2.h"
#include "ext2_volume.h"
#include "ext2_file.h"

rf_int_t ext2_volume_open(rf_ext2_volume_ref_t vol, rf_mount_param_t* param)
{
    rf_u32_t sector_bytes;
    rf_u32_t rev_level;
    rf_u32_t blocks_per_group;
    rf_u32_t s_feature_incompat, s_feature_compat;
    rf_int_t ret;

    ret = rf_disk_open(param->device_name, param->flags, param->ctx, &vol->dev_handle);
    if (ret)
        return ret;
    vol->access_flag = 0;
    if ((param->flags & 0x3) == 2)
        vol->access_flag = 1;

    ret = rf_disk_ioctl(vol->dev_handle, RF_CMD_GET_SECTOR_SIZE, &sector_bytes, sizeof(sector_bytes));
    if (ret)
        return ret;
    RF_PERROR("sector bytes error", INVOID_SECTOR_SIZE(sector_bytes), return RF_ENOFS);

#if RF_OPT_USE_MALLOC == 1
    vol->sector_buffer = malloc(sector_bytes);
    if (vol->sector_buffer == NULL)
        return RF_ENOMEM;
#else
    if (RF_OPT_MAX_SECTOR_SIZE < sector_bytes)
        return RF_EBUFSMALL;
#endif

    ret = rf_disk_read(vol->dev_handle, 2, 1, vol->sector_buffer);
    if (ret)
        return ret;
    /* MAGIC жǷextļϵͳ */
    if (vol->sector_buffer[0x38] != 0x53 || vol->sector_buffer[0x39] != 0xef)
        return RF_EPERM;
    vol->sector_size = (rf_u16_t)sector_bytes;
    vol->sector_size_bits = rf_ffs(vol->sector_size);
    vol->blocks_count = u32_read(vol->sector_buffer, 0x04);
    vol->free_blocks_count = u32_read(vol->sector_buffer, 0x0c);
    vol->block_size = EXT2_MIN_BLOCK_SIZE << u32_read(vol->sector_buffer, 0x18);
    vol->block_size_bits = rf_ffs(vol->block_size);
    vol->sectors_per_cluster = (rf_u8_t)(vol->block_size >> vol->sector_size_bits);
    vol->sectors_per_cluster_bits = rf_ffs(vol->sectors_per_cluster);
    vol->inodes_per_group = u32_read(vol->sector_buffer, 0x28);

    rev_level = u32_read(vol->sector_buffer, 0x4c);
    if (rev_level == EXT2_GOOD_OLD_REV)
        vol->inode_size = EXT2_GOOD_OLD_INODE_SIZE;
    else
        vol->inode_size = u16_read(vol->sector_buffer, 0x58);

    blocks_per_group = u32_read(vol->sector_buffer, 0x20);
    if (blocks_per_group == 0)
        return RF_ERROR;
    vol->total_groups = u32_read(vol->sector_buffer, 0x4) / blocks_per_group;

    vol->group_start_sector = vol->block_size >> vol->sector_size_bits;
    if (vol->group_start_sector <= 2)
        vol->group_start_sector += 2;

    vol->fs_type = RF_TYPE_EXT2;
    s_feature_compat = u32_read(vol->sector_buffer, 0x5c);
    s_feature_incompat = u32_read(vol->sector_buffer, 0x60);
    if ((s_feature_incompat & EXT3_FEATURE_INCOMPAT_JOURNAL_DEV) ||
            (s_feature_incompat & EXT3_FEATURE_INCOMPAT_RECOVER) ||
            (s_feature_compat & EXT3_FEATURE_COMPAT_HAS_JOURNAL)) {
        vol->fs_type = RF_TYPE_EXT3;
    }

    ret = ext2_get_inode_entry(vol, 2, &vol->root_dir_entry);
    if (ret != 0)
        return ret;
    /* ȡĿ¼ڵ ΪǰĿ¼ */
    vol->current_dir_entry.sector = vol->root_dir_entry.sector;
    vol->current_dir_entry.offset = vol->root_dir_entry.offset;
    return 0;
}

rf_int_t ext2_volume_close(rf_ext2_volume_ref_t vol)
{
    rf_int_t ret;

    ret = rf_disk_close(vol->dev_handle);
    if (ret == RF_ESUCCESS) {
#if RF_OPT_USE_MALLOC == 1
        if (vol->sector_buffer != NULL)
            rf_free(vol->sector_buffer);
#endif
    }
    return ret;
}

rf_int_t ext2_volume_stat(rf_ext2_volume_ref_t vol, rf_statfs_ref_t st)
{
    st->f_type = vol->fs_type;
    st->f_ssize = vol->sector_size;
    st->f_bsize = vol->sectors_per_cluster << vol->sectors_per_cluster_bits;
    st->f_bfree = vol->free_blocks_count;
    st->f_blocks = vol->blocks_count;
    return 0;
}

rf_int_t ext2_volume_get_label(rf_ext2_volume_ref_t vol, const rf_char_t* buf, rf_int_t len)
{
    return 0;
}

rf_int_t ext2_volume_set_label(rf_ext2_volume_ref_t vol, const rf_char_t* buf)
{
    return 0;
}

rf_int_t ext2_volume_set_current_dir(rf_ext2_volume_ref_t vol, rf_fentry_t* entry)
{
    vol->current_dir_entry.sector = entry->sector;
    vol->current_dir_entry.offset = entry->offset;
    return 0;
}

