/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#define _RF_MAX_NAME 520
#include "../common/rf_common.h"
#include "fat_volume.h"
#include "fat_file.h"
#include "fat.h"

#define DIR_Name            0       /* Short file name (11) */
#define DIR_Attr            11      /* Attribute (1) */
#define DIR_NTres           12      /* Lower case flag (1) */
#define DIR_CrtTimeTenth    13      /* Created time sub-second (1) */
#define DIR_CrtTime         14      /* Created time (2) */
#define DIR_CrtDate         16      /* Created date (2) */
#define DIR_LstAccDate      18      /* Last accessed date (2) */
#define DIR_FstClusHI       20      /* Higher 16-bit of first cluster (2) */
#define DIR_WrtTime         22      /* Modified time (2) */
#define DIR_WrtDate         24      /* Modified date (2) */
#define DIR_FstClusLO       26      /* Lower 16-bit of first cluster (2) */
#define DIR_FileSize        28      /* File size (4) */

#define XDIR_Type           0       /* exFAT: Type of exFAT directory entry (BYTE) */
#define XDIR_NumLabel       1       /* exFAT: Number of volume label characters (BYTE) */
#define XDIR_Label          2       /* exFAT: Volume label (11-WORD) */
#define XDIR_CaseSum        4       /* exFAT: Sum of case conversion table (DWORD) */
#define XDIR_NumSec         1       /* exFAT: Number of secondary entries (BYTE) */
#define XDIR_SetSum         2       /* exFAT: Sum of the set of directory entries (WORD) */
#define XDIR_Attr           4       /* exFAT: File attribute (WORD) */
#define XDIR_CrtTime        8       /* exFAT: Created time (DWORD) */
#define XDIR_ModTime        12      /* exFAT: Modified time (DWORD) */
#define XDIR_AccTime        16      /* exFAT: Last accessed time (DWORD) */
#define XDIR_CrtTime10      20      /* exFAT: Created time subsecond (BYTE) */
#define XDIR_ModTime10      21      /* exFAT: Modified time subsecond (BYTE) */
#define XDIR_CrtTZ          22      /* exFAT: Created timezone (BYTE) */
#define XDIR_ModTZ          23      /* exFAT: Modified timezone (BYTE) */
#define XDIR_AccTZ          24      /* exFAT: Last accessed timezone (BYTE) */
#define XDIR_GenFlags       33      /* exFAT: General secondary flags (BYTE) */
#define XDIR_NumName        35      /* exFAT: Number of file name characters (BYTE) */
#define XDIR_NameHash       36      /* exFAT: Hash of file name (WORD) */
#define XDIR_ValidFileSize  40      /* exFAT: Valid file size (QWORD) */
#define XDIR_FstClus        52      /* exFAT: First cluster of the file data (DWORD) */
#define XDIR_FileSize       56      /* exFAT: File/Directory size (QWORD) */

#define SZDIRE              32      /* Size of a directory entry */

#define  RF_FILE_CREATE_FLAG    1
#define  RF_FILE_WRITE_FLAG     2

static rf_int_t fat_file_update_time(rf_fat_file_ref_t file, rf_int_t flags);

rf_int_t fat_file_write_long_name(rf_fat_file_ref_t pdir, rf_fat_file_ref_t pff, rf_u32_t offset,
                                  rf_u32_t sum, const rf_char_t* longname);

static rf_int_t fat_file_open_end_check(rf_fat_file_ref_t file, rf_u32_t flag)
{
    rf_int_t ret;

    if (file->attribute & RF_ATTR_DIRECTORY)
        return RF_ESUCCESS;
    /* ȷҪ Ŀ¼ 򿪵ȴļ ش */
    if (flag & RF_O_DIRECTORY)
        return RF_ENOTDIR;
    if (flag & RF_O_APPEND)
        return fat_file_lseek(file, 0, RF_SEEK_END);
    if (flag & RF_O_TRUNC) {
#if RF_OPT_FS_READONLY == 0
#if RF_OPT_USE_FTRUNCATE
        ret =  fat_file_truncate(file, 0);
        if (ret)
            return ret;
        /* ļ  дʱ */
        ret = fat_file_sync(file);
        if (ret)
            return ret;
#else
        return RF_EPERM;
#endif
#else
        return RF_EROFS;
#endif
    }
    return RF_ESUCCESS;
}

static rf_int_t fat_file_raw_open_imp(rf_fat_file_ref_t file, rf_fentry_t* fentry,
                                      rf_u32_t up_dir_start_clst, rf_u32_t up_dir_offset, rf_u32_t open_flag)
{
    rf_int_t ret, len;
    rf_u8_t* entry;

    file->access_flag = (rf_u8_t)open_flag & 0x03;
    file->entry.sector = fentry->sector;
    file->entry.offset = fentry->offset;

#if RF_OPT_FILE_ACCESS_CHECK
    if (!fat_check_file_access(file))
        return RF_EACCES;
#endif

    /* 򿪵 ǰĿ¼ */
    chain_init(&file->clust, TRUE);
    file->clust.chain_offset = 1;
    file->gather_all_clust_bit = FALSE;
    file->sector_buf_flag = FALSE;
    file->first_gather_cluester_bit = TRUE;
    file->root_dir_flag = FALSE;

    if (fentry->sector == 0) {
_OPEN_ROOT_DIR:
        file->root_dir_flag = TRUE;
        file->attribute = RF_ATTR_DIRECTORY;
        file->start_cluster = 0;
        file->updir_start_cluster = 0;

        if (file->vol->fs_type == RF_TYPE_FAT32) {
            file->start_cluster = file->vol->bpb_root_start_sector;
            /*  fat32 Ŀ¼ Ѽش  */
#if RF_OPT_FILE_SECBUF
            if (open_flag & RF_O_SECBUF) {
                file->sector_buf_flag = TRUE;
                return clusters_gather_file(file, TRUE);
            }
#endif
            return RF_ESUCCESS;
        } else {
            /* Ŀ¼ļ ȡļƫ */
            file->current_offset = 0;
            /* ȡƫ */
            file->size = 0;
            /* clust.byte_offset Ŀ¼ȡƫ */
            file->clust.byte_offset = 0;
            return RF_ESUCCESS;
        }
    } else {
        ret = rf_disk_read(file->vol->dev_handle, fentry->sector, 1, file->vol->sector_buffer);
        if (ret)
            return ret;

        entry = file->vol->sector_buffer + fentry->offset;
        file->start_cluster = u16_read(entry, DIR_FstClusHI) << 16 | u16_read(entry, DIR_FstClusLO);
        file->attribute = u8_read(entry, DIR_Attr);
        file->current_offset = 0;
        file->size = u32_read(entry, DIR_FileSize);
        if (file->attribute & RF_ATTR_DIRECTORY) {
            /*Ŀ¼ֻֻ*/
            if (file->start_cluster == 0)
                goto _OPEN_ROOT_DIR;
        }

#if RF_OPT_FILE_SECBUF
        if (open_flag & RF_O_SECBUF) {
            file->sector_buf_flag = TRUE;
            if (file->start_cluster == 0) {
                ret = clusters_file_build_end(file, 1, (rf_u32_t*)&len);
                if (RF_ESUCCESS != ret)
                    return ret;
                if (len == 0)
                    return RF_ENOSPC;
            } else {
                ret = clusters_gather_file(file, TRUE);
                if (RF_ESUCCESS != ret)
                    return ret;
            }
        }
#endif
        /* file->start_cluster == 0xffffffff Ŀ¼ʱʱʹ */
        if (file->start_cluster == 0 ||  file->start_cluster == 0xffffffff) {
            file->gather_all_clust_bit = TRUE;
        }

        file->updir_start_cluster = up_dir_start_clst;
        file->up_dir_offset = up_dir_offset;
    }
    return RF_ESUCCESS;
}

#if RF_OPT_USE_RAWOPEN
/* ļƫ ϼĿ¼غƫ ٴļ */
rf_int_t fat_file_raw_open(rf_fat_file_ref_t file, rf_fentry_t* fentry, rf_int_t flag)
{
    rf_int_t ret;

    ret = fat_file_raw_open_imp(file, fentry, -1, 0, flag);
    if (ret)
        return ret;
    return fat_file_open_end_check(file, flag);
}
#endif

#if RF_OPT_USE_SET_NAME == 1 || RF_OPT_USE_GET_SET_LONG_NAME == 1
static rf_int_t fat_dir_open(rf_fat_file_ref_t dir, rf_u32_t flag)
{
#if RF_OPT_FILE_ACCESS_CHECK
    if (!check_start_cluster_access(dir->vol, dir->start_cluster, (rf_bool_t)(flag & 0x3)))
        return RF_EACCES;
#endif

    dir->open_flag = 1;
    dir->attribute = RF_ATTR_DIRECTORY;
    dir->current_offset = 0;
    dir->first_gather_cluester_bit = FALSE;
    return clusters_gather_file(dir, TRUE);
}
#endif

/*
name ԭʼļַ
name Ϊnull ƥļ
name  *.txt Ҫļת
*/
rf_int_t fat_file_read_dir(rf_fat_file_ref_t dir, const rf_char_t* name, rf_dirent_ref_t pfd, rf_int_t flag)
{
    rf_int_t ret;
    rf_char_t* entry;
    rf_u32_t offset, len;

#if RF_OPT_FAT_LONG_NAME
    rf_wchar_t* longname = (rf_wchar_t*)pfd->name;
    rf_u32_t i, k, l_offset, sum = 0;
    rf_u32_t last_offset = 0;
#endif

    if (!(dir->attribute & RF_ATTR_DIRECTORY))
        return RF_ENOTDIR;
    offset = (rf_u32_t)(dir->current_offset) % dir->vol->sector_size;
    dir->current_offset -= offset;
    do {
        ret = fat_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &len);
        if (ret != RF_ESUCCESS)
            return ret;
        if (!len) {
#if RF_OPT_FS_READONLY == 0
            dir->dir_end_clst_bit = TRUE;
#endif
            return RF_ENDFILE;
        }

        for (; offset < dir->vol->sector_size; offset += 32) {
            entry = (rf_char_t*)dir->vol->sector_buffer + offset;
            if (*(entry + DIR_Name) == FILE_EMPTY) {
                dir->current_offset = dir->current_offset - dir->vol->sector_size + offset;
                return RF_ENDFILE;
            }
            if ((rf_u8_t) * (entry + DIR_Name) == FILE_DELETED)
                continue;

            /* readdir Ҫȡļ  fopen򿪳ļ */
            if (*(entry + DIR_Attr) == RF_ATTR_LFN_ENTRY) {
                if ((pfd->name != NULL) && (!(flag & RF_O_READDIR_SFN) || flag & RF_O_LONGNAME)) {
#if RF_OPT_FAT_LONG_NAME
                    k = 0;
                    l_offset = ((u8_read(dir->vol->sector_buffer, offset) & 0XBF) - 1) * 13;
                    if (l_offset == 0)
                        last_offset = dir->current_offset - dir->vol->sector_size + offset;
                    if ((l_offset + 13) * 2 > pfd->len)
                        return RF_EBUFSMALL;

                    /*ڴUNICODEС˷ʽ*/
                    for (i = 0; i < 5; i++)
                        longname[l_offset + k++] = u16_read(dir->vol->sector_buffer, offset + i * 2 + 1);
                    for (i = 0; i < 6; i++)
                        longname[l_offset + k++] = u16_read(dir->vol->sector_buffer, offset + 14 + i * 2);
                    for (i = 0; i < 2; i++)
                        longname[l_offset + k++] = u16_read(dir->vol->sector_buffer, offset + 28 + i * 2);

                    sum = u8_read(dir->vol->sector_buffer, offset + 13);
#else
                    return RF_EPERM;
#endif
                }
                continue;
            }

            /* ļֿ 1 ԭʼļ 2ļ ƥ
                ģʽƥ 3ļģʽƥ 4ļģʽƥ */
            if (flag & RF_O_RAW_SFN) {
                /* ƥַ ֱӱȽ */
                if (memcmp(entry, name, 11))
                    continue;
            } else if (flag & RF_O_LONGNAME) {
#if RF_OPT_FAT_LONG_NAME
                /* 򿪳ļ· */
                if (rf_wcscmp((rf_wchar_t*)longname, (rf_wchar_t*)name))
                    continue;
#else
                return RF_EPERM;
#endif
            } else if (flag & RF_O_READDIR_SFN) {
                if (pfd->name) {
                    /* readdir ģʽƥ */
                    fat_convert_raw_name(entry, *(entry + 12), pfd->name);
                    if (!rf_str_in_pattern(pfd->name, name))
                        continue;
                }
            } else {
                /* ļ ʱ֧ ģʽƥ䣡 */
#if RF_OPT_READDIR_LONG_NAME

#endif
            }

            dir->current_offset = dir->current_offset - dir->vol->sector_size + offset + 32;
            if (pfd) {
                pfd->entry.sector = dir->vol->current_work_sector;
                pfd->entry.offset = (rf_u16_t)offset;
                pfd->attribute = *(entry + DIR_Attr);
#if RF_OPT_FAT_LONG_NAME
                /* ۼӺ  ļƫ Ҫ ʹöļ  */
                if ((!(flag & RF_O_READDIR_SFN))
                        && sum == fat_long_file_name_sum(dir->vol->sector_buffer + offset) && (last_offset + 64 == dir->current_offset)) {
                    pfd->type = RF_CODE_PAGE_UCS2L;
                    goto _EXIT;
                }
#endif
                pfd->type = RF_NOW_CODE_PAGE;
                if (pfd->name)
                    fat_convert_raw_name(entry, *(entry + 12), pfd->name);
            }
#if RF_OPT_FAT_LONG_NAME
_EXIT:
#endif
            if (dir->current_offset % dir->vol->sector_size != 0)
                clusters_back_offset_sector(dir, 1);
            return RF_ESUCCESS;
        }
        offset = 0;
    } while (1);
}

#if RF_OPT_FAT_LONG_NAME
/* Ŀ¼ Ҷļ ڳļĴ */
rf_int_t fat_file_find_sfn(rf_fat_file_ref_t file, const rf_char_t* sfn, rf_u32_t open_flag)
{
    rf_int_t ret;
    rf_dirent_t fdir;

    memset(&fdir, 0, sizeof(fdir));
    file->current_offset = 0;
    ret = fat_file_lseek(file, 0, RF_SEEK_SET);
    if (ret)
        return ret;
    return fat_file_read_dir(file, sfn, &fdir, open_flag | RF_O_RAW_SFN);
}
#endif

static rf_int_t fat_file_open_imp(rf_fat_file_ref_t file, const rf_char_t* sfn, rf_wchar_t* lfn, rf_u32_t open_flag)
{
    rf_int_t ret;
    rf_dirent_t fdir;
    const rf_char_t* name;
#if RF_OPT_FAT_LONG_NAME
    DEF_FIND_BUF
#endif

    memset(&fdir, 0, sizeof(fdir));
    name = sfn;
    if (*((rf_u8_t*)&open_flag + 2) & NS_LFN) {
        /* Ҫ RF_OPT_FAT_LONG_NAME */
#if RF_OPT_FAT_LONG_NAME == 0
        return RF_EPERM;
#endif
        name = (const rf_char_t*)lfn;
        open_flag |= RF_O_LONGNAME;
    } else {
        open_flag |= RF_O_RAW_SFN;
    }

#if RF_OPT_FAT_LONG_NAME
    GET_FIND_BUF(fdir.name);
#endif
    fdir.len = _RF_MAX_NAME;
    ret = fat_file_read_dir(file, name, &fdir, open_flag);
#if RF_OPT_FAT_LONG_NAME
    FREE_FIND_BUF(fdir.name);
#endif
    if (ret != RF_ESUCCESS)
        return ret;

    file->entry.sector = fdir.entry.sector;
    file->entry.offset = fdir.entry.offset;
    return fat_file_raw_open_imp(file, &file->entry, file->start_cluster, file->current_offset - 0x20, open_flag);
}

#if RF_OPT_FS_READONLY == 0
#define TM2FT( tm ) ( ((rf_u32_t)(tm.tm_year - 1980) << 25) \
        | ((rf_u32_t)tm.tm_mon << 21) \
        | ((rf_u32_t)tm.tm_mday << 16) \
        | ((rf_u32_t)tm.tm_hour << 11) \
        | ((rf_u32_t)tm.tm_min << 5) \
        | ((rf_u32_t)tm.tm_sec >> 1) )
rf_u32_t fat_time_current(rf_void_t)
{
    rf_tm_t tm;
    if (rf_getlocaltime(&tm) != 0)
        return 0;
    return TM2FT(tm);
}

static rf_int_t fat_dir_add_clst(rf_fat_file_ref_t dir)
{
    rf_int_t ret;
    rf_u32_t tmp, count;

    /* ´ ҳʼ0 */
    ret = clusters_file_build_end(dir, 1, &tmp);
    if (ret)
        return ret;
    if (!tmp)
        return RF_ENOSPC;
    memset(dir->vol->sector_buffer, 0, dir->vol->sector_size);
    for (tmp = 0; tmp < dir->vol->sectors_per_cluster; tmp++) {
        ret = clusters_file_write_sector(dir, dir->vol->sector_buffer, 1, &count);
        if (ret)
            return ret;
    }
    clusters_back_offset_sector(dir, dir->vol->sectors_per_cluster);
    dir->vol->current_work_sector = dir->vol->fat_data_start_sector + ((dir->clust.chain[0].cluster - 2) << dir->vol->sectors_per_cluster_bits);
    ret = rf_disk_read(dir->vol->dev_handle, dir->vol->current_work_sector, 1, dir->vol->sector_buffer);
    if (ret)
        return ret;
    dir->clust.byte_offset = 0;
    return RF_ESUCCESS;
}

#if 0
static rf_u16_t xdir_sum(   /* Get checksum of the directoly entry block */
    const rf_u8_t* dir      /* Directory entry block to be calculated */
)
{
    rf_int_t i, szblk;
    rf_u16_t sum;

    szblk = (dir[XDIR_NumSec] + 1) * SZDIRE;
    for (i = sum = 0; i < szblk; i++) {
        if (i == XDIR_SetSum) { /* Skip sum field */
            i++;
        } else {
            sum = ((sum & 1) ? 0x8000 : 0) + (sum >> 1) + dir[i];
        }
    }
    return sum;
}

static rf_u16_t xname_sum(  /* Get check sum (to be used as hash) of the name */
    const rf_wchar_t* name  /* File name to be calculated */
)
{
    rf_wchar_t chr;
    rf_u16_t sum = 0;

    while ((chr = *name++) != 0) {
        chr = wtoupper(chr);        /* File name needs to be upper-case converted */
        sum = ((sum & 1) ? 0x8000 : 0) + (sum >> 1) + (chr & 0xFF);
        sum = ((sum & 1) ? 0x8000 : 0) + (sum >> 1) + (chr >> 8);
    }
    return sum;
}
#endif


static rf_int_t fat_file_create(rf_fat_file_ref_t dir, const rf_char_t* sfn, const rf_wchar_t* lfn,  rf_u32_t open_flag)
{
    rf_int_t ret;
    rf_u32_t updir_cluster, offset;
    rf_char_t* entry;
    rf_u8_t ns = *(sfn + NS);

    ns = *(sfn + NS);
    if (dir->dir_end_clst_bit) {
        ret = fat_dir_add_clst(dir);
        if (ret)
            return ret;
        dir->dir_end_clst_bit = FALSE;
    }

    updir_cluster = dir->start_cluster;
    offset = dir->current_offset % dir->vol->sector_size;
    if (*((rf_u8_t*)&open_flag + 2) & NS_LFN) {
#if RF_OPT_FAT_LONG_NAME
        ret = fat_file_write_long_name(dir, dir, offset, fat_long_file_name_sum((rf_u8_t*)sfn), (const rf_char_t*)lfn);
        if (ret)
            return ret;
        offset = dir->entry.offset;
#else
        return RF_EPERM;
#endif
    }

    /* ʼļ */
    entry = (rf_char_t*)(dir->vol->sector_buffer + offset);
    memset(entry, 0, 32);
    memcpy(entry, sfn, 11);
    /* ļʱ */
    u32_write(entry, DIR_CrtTime, fat_time_current());
    u16_write(entry, DIR_LstAccDate, (rf_u16_t)fat_time_current());
    entry[DIR_NTres] = *((rf_u8_t*)&open_flag + 2);
    if (open_flag & RF_O_DIRECTORY) {
        /* Ŀ¼ʼ ΪЧʼ 0 */
        u16_write(entry, DIR_FstClusHI, 0xffff);
        u16_write(entry, DIR_FstClusLO, 0xffff);
        u8_write(entry, DIR_Attr, RF_ATTR_DIRECTORY);
    }
    ret = rf_disk_write(dir->vol->dev_handle, dir->vol->current_work_sector, 1, dir->vol->sector_buffer);
    if (ret)
        return ret;

    /* ӦøĿ¼дʱ */
    ret = fat_file_update_time(dir, RF_FILE_WRITE_FLAG);
    if (ret)
        return ret;

    /* ʹõǰ ǰļ */
    dir->entry.sector = dir->vol->current_work_sector;
    dir->entry.offset = (rf_u16_t)offset;
    ret = fat_file_raw_open_imp(dir, &dir->entry, dir->start_cluster, dir->current_offset, open_flag);
    if (ret)
        return ret;
    if (open_flag & RF_O_DIRECTORY) {
        dir->start_cluster = 0;
        ret = fat_dir_add_clst(dir);
        if (ret)
            return ret;

        strcpy((rf_char_t*)dir->vol->sector_buffer, ".          ");
        dir->vol->sector_buffer[DIR_Attr] = 0x10;
        u16_write(dir->vol->sector_buffer, DIR_FstClusHI, HIWORD(dir->start_cluster));
        u16_write(dir->vol->sector_buffer, DIR_FstClusLO, LOWORD(dir->start_cluster));

        strcpy((rf_char_t*)dir->vol->sector_buffer + 32, "..         ");
        dir->vol->sector_buffer[32 + DIR_Attr] = 0x10;

        /* fat32 .. ָĿ¼  Ϊ 0 ⴦ */
        if (updir_cluster == 2 && dir->vol->fs_type == RF_TYPE_FAT32)
            updir_cluster = 0;
        u16_write(dir->vol->sector_buffer + 32, DIR_FstClusHI, HIWORD(updir_cluster));
        u16_write(dir->vol->sector_buffer + 32, DIR_FstClusLO, LOWORD(updir_cluster));

        ret = rf_disk_write(dir->vol->dev_handle, dir->vol->current_work_sector, 1, dir->vol->sector_buffer);
        if (ret)
            return ret;

        /* Ŀ¼ʼ */
        return fat_file_sync(dir);
    }
    return RF_ESUCCESS;
}

static rf_int_t fat_dir_create_file(rf_fat_file_ref_t file, const rf_char_t* sfn, const rf_wchar_t* lfn, rf_u32_t flag)
{
#if RF_OPT_FAT_LONG_NAME
    rf_char_t fn[12];
    rf_int_t ret, n;
#endif
    rf_int_t ns;

    file->access_flag = RF_O_RDWR;
#if RF_OPT_FILE_ACCESS_CHECK
    if (!fat_check_file_access(file))
        return RF_EACCES;
#endif

    ns = *(sfn + NS);
#if (RF_OPT_FAT_LONG_NAME==0)
    if (ns & NS_LOSS)
        return RF_EPERM;
    return fat_file_create(file, sfn, lfn, flag);
#else
    /* ڵǰĿ¼ ļǷ!!! */
    if (ns & NS_LOSS) {
        for (n = 1; n < 100; n++) {
            fat_gen_numname(fn, sfn, lfn, n);
            ret = fat_file_find_sfn(file, fn, flag);
            /* ûҵ ˵ļ */
            if (ret == RF_ENDFILE)
                break;
            if (ret != RF_ESUCCESS)
                return ret;
        }
        if (n == 100)
            return RF_EPERM;
        return fat_file_create(file, fn, lfn, flag);
    } else {
        return fat_file_create(file, sfn, lfn, flag);
    }
#endif
}
#endif

rf_int_t fat_file_open(rf_fat_file_ref_t file, const rf_char_t* path, rf_int_t flag)
{
    rf_int_t ret, ns = 0;
    rf_fentry_t fentry;
    rf_u32_t updir_start_clst, updir_offset;
#if RF_OPT_FAT_LONG_NAME
    DEF_PATH_BUF
#endif
    rf_char_t* lfn = NULL;
    rf_char_t sfn[12];

    file->access_flag = flag & 0x3;
    /* ֻļϵͳ */
    if (file->access_flag == 1 && file->vol->access_flag == 0)
        return RF_EROFS;

    /*  ֧· 򿪷ǰĿ¼ */
    if (path[0] == '\0' || is_absolute_path(path)) {
        fentry.sector = 0;
        fentry.offset = 0;
        updir_start_clst = 0;
        updir_offset = 0;
        if (path[0] != '\0')
            path += 1;
    } else {
        fentry = file->vol->current_dir_entry;
        updir_start_clst = file->vol->updir_start_clst;
        updir_offset = file->vol->updir_offset;
    }
    ret = fat_file_raw_open_imp(file, &fentry, updir_start_clst, updir_offset, RF_O_RDONLY);
    if (ret != RF_ESUCCESS)
        return ret;

#if RF_OPT_FAT_LONG_NAME
    GET_PATH_BUF(lfn);
#endif
    while (!(ns & NS_LAST)) {
        ret = path_get_sfn_lfn(&path, (rf_char_t*)sfn, (rf_wchar_t*)lfn);
        if (ret != RF_ESUCCESS) {
            /* 򿪸Ŀ¼ 0:/ */
            if (ret == -1)
                ret = 0;
            break;
        }
        if (sfn[0] == '\0')
            break;

        ns = *(sfn + NS);
        /* nsflag ĵ3ֽ */
        *((rf_u8_t*)&flag + 2) = (rf_u8_t)ns;
        *(sfn + NS) = '\0';

        if (ns & NS_DOT)
            if (!memcmp(sfn, ". ", 2))
                continue;
        if (ns & NS_LAST) {
            ret = fat_file_open_imp(file, sfn, (rf_wchar_t*)lfn, flag);
            if (ret == RF_ENDFILE) {
                /* ڸĿ¼ ϼĿ¼*/
                if ((ns & NS_DOT) && !memcmp(sfn, ".. ", 3))
                    ret = RF_ESUCCESS;
            }
        } else {
            /* мΪĿ¼ ʹֻ ȥ RF_O_APPEND RF_O_TRUNC ־   */
            ret = fat_file_open_imp(file, sfn, (rf_wchar_t*)lfn, flag & (~0xf));
        }

        if (ret != RF_ESUCCESS) {
            /* һĿ¼ ļĿ¼ */
            if (ret == RF_ENDFILE && ns & NS_LAST && flag & RF_O_CREAT) {
                /* һĿ¼ Ƿд */
#if RF_OPT_FS_READONLY == 0
                *(sfn + NS) = ns;
                ret = fat_dir_create_file(file, sfn, (rf_wchar_t*)lfn, flag);
#else
                ret = RF_EROFS;
#endif
            }
            if (ret)
                break;
        }

        /* ֻһ ļ  */
        if (!(ns & NS_LAST) && !(file->attribute & RF_ATTR_DIRECTORY)) {
            RF_PERROR("Path to the include file error\n", 1, ret = RF_ENOTDIR);
            break;
        }
    }
#if RF_OPT_FAT_LONG_NAME
    FREE_PATH_BUF(lfn);
#endif
    if (ret == RF_ENDFILE)
        ret = RF_ENOENT;
    if (ret != RF_ESUCCESS)
        return ret;
    return fat_file_open_end_check(file, flag);
}

rf_int_t fat_file_close(rf_fat_file_ref_t file)
{
#if RF_OPT_FS_READONLY == 0
    rf_int_t ret;
    ret = fat_file_sync(file);
    if (ret)
        return ret;
#endif
    file->open_flag = FALSE;
    return RF_ESUCCESS;
}

#if RF_OPT_USE_GET_FNAME
rf_int_t fat_file_get_name(rf_fat_file_ref_t file, rf_char_t* long_name, rf_u32_t len)
{
    rf_int_t ret;
    rf_char_t tmp[13];
    rf_u8_t nt;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    nt = *(file->vol->sector_buffer + file->entry.offset + 12);
    fat_convert_raw_name((rf_char_t*)file->vol->sector_buffer + file->entry.offset, nt, tmp);

    if (len < strlen(tmp) + 1)
        return RF_EBUFSMALL;
    strcpy(long_name, tmp);
    return RF_ESUCCESS;
}
#endif

/* Ҫ򿪳ļ֧ */
#if RF_OPT_USE_GET_SET_LONG_NAME
static rf_int_t fat_file_long_name(rf_fat_file_ref_t file, rf_wchar_t* long_name, rf_u32_t len, rf_bool_t rm_name_flag)
{
    rf_int_t ret, i, j, k;
    rf_u8_t sum;
    rf_u32_t read_len, offset, sector_offset;
    rf_fat_file_t tdir;
    rf_bool_t need_write_flag;
    rf_bool_t fln_bit;

    /* Ŀ¼ʼƫ ûгļ */
    if (file->up_dir_offset == 0)
        return RF_ESUCCESS;

    if (long_name)
        long_name[0] = 0;
    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    sum = fat_long_file_name_sum(file->vol->sector_buffer + file->entry.offset);
    memset(&tdir, 0, sizeof(tdir));
    tdir.vol = file->vol;
    tdir.start_cluster = file->updir_start_cluster;
    ret = fat_dir_open(&tdir, RF_O_RDONLY);
    if (ret != RF_ESUCCESS)
        return ret;

    /* 640 = 20 * 32 ļƫ */
    offset = file->up_dir_offset;
    if (offset <= 640)
        offset = 0;
    else
        offset -= 640;

    sector_offset = offset % file->vol->sector_size;
    ret = fat_file_lseek(&tdir, offset - sector_offset, RF_SEEK_SET);
    if (ret != RF_ESUCCESS)
        return ret;

    ret = fat_file_read(&tdir, file->vol->sector_buffer, file->vol->sector_size, &read_len);
    if (ret != RF_ESUCCESS)
        return ret;

    /* ֹѭ */
    need_write_flag = FALSE;
    fln_bit = FALSE;
    for (j = 0; j <= 20; j++, sector_offset += 32) {
        if (tdir.current_offset - tdir.vol->sector_size + sector_offset == file->up_dir_offset)
            break;

        if (sector_offset == file->vol->sector_size) {
            if (need_write_flag) {
                need_write_flag = FALSE;
                ret = rf_disk_write(file->vol->dev_handle, tdir.vol->current_work_sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
            }
            ret = fat_file_read(&tdir, file->vol->sector_buffer, file->vol->sector_size, &read_len);
            if (ret != RF_ESUCCESS)
                return ret;
            sector_offset = 0;
        }

        if ((u8_read(file->vol->sector_buffer, sector_offset + 11) == 0xf)
                && (sum == u8_read(file->vol->sector_buffer, sector_offset + 13))
                && (u8_read(file->vol->sector_buffer, sector_offset) & 0x40)) {
            offset = (u8_read(file->vol->sector_buffer, sector_offset) & 0XBF) * 32;
            if (tdir.current_offset - tdir.vol->sector_size + sector_offset +  offset == file->up_dir_offset)
                fln_bit = TRUE;
        }

        if (!fln_bit)
            continue;
        if (rm_name_flag) {
            file->vol->sector_buffer[sector_offset] = 0xe5;
            need_write_flag = TRUE;
        } else {
            k = 0;
            offset = ((u8_read(file->vol->sector_buffer, sector_offset) & 0XBF) - 1) * 26;
            for (i = 0; i < 5; i++)
                long_name[offset + k++] = u16_read(file->vol->sector_buffer, sector_offset + i * 2 + 1);
            for (i = 0; i < 6; i++)
                long_name[offset + k++] = u16_read(file->vol->sector_buffer, sector_offset + 14 + i * 2);
            for (i = 0; i < 2; i++)
                long_name[offset + k++] = u16_read(file->vol->sector_buffer, sector_offset + 28 + i * 2);

            if (u8_read(file->vol->sector_buffer, sector_offset) & 0X40) {
                /*պΪ26ֽڳļ*/
                if (!((long_name[offset + k - 1] == 0x00) || ((rf_u16_t)long_name[offset + k - 1] == 0xffff))) {
                    if (offset + 28 > len)
                        return RF_EBUFSMALL;
                    long_name[offset + 13] = 0x00;
                }
            }
        }
    }
    if (need_write_flag) {
        need_write_flag = FALSE;
        ret = rf_disk_write(file->vol->dev_handle, tdir.vol->current_work_sector, 1, file->vol->sector_buffer);
        if (ret)
            return ret;

    }
    return RF_ESUCCESS;
}

rf_int_t fat_file_get_long_name(rf_fat_file_ref_t file, rf_wchar_t* long_name, rf_u32_t len)
{
    return fat_file_long_name(file, long_name, len, FALSE);
}

rf_int_t fat_file_rm_long_name(rf_fat_file_ref_t file)
{
    return fat_file_long_name(file, NULL, 0, TRUE);
}

rf_int_t fat_file_write_long_name(rf_fat_file_ref_t dir, rf_fat_file_ref_t file, rf_u32_t offset, rf_u32_t sum, const rf_char_t* long_name)
{
    rf_u32_t ret, read_len, len;
    rf_u32_t i, j, k, lc, lo;
    rf_wchar_t* tmp;

    /* unicodeļ㳤!!! */
    len = (rf_int_t)(rf_wstrlen((const rf_wchar_t*)long_name) + 1) * 2;
    /* һĿ¼0xff */
    memset(dir->vol->sector_buffer + offset, 0xff, 32);
    lc = (len + 26 - 1) / 26;
    lo = len % 26;
    if (lo == 0)
        lo = 26;
    for (i = lc; i > 0; i--) {
        k = 0;
        if (lo) {
            dir->vol->sector_buffer[offset + 0] = (rf_u8_t)i;
            if (i == lc)
                dir->vol->sector_buffer[offset + 0] |= 0x40;

            tmp = (rf_wchar_t*)dir->vol->sector_buffer;
            dir->vol->sector_buffer[offset + 13] = sum;
            u16_write(tmp, offset + 11, 0x0f);
            u16_write(tmp, offset + 26, 0x00);
            for (j = 0; j < 5; j++) {
                if (k < lo)
                    u16_write(dir->vol->sector_buffer, offset + j * 2 + 1, *(rf_u16_t*)(long_name + ((i - 1) * 26) + k));
                k += 2;
            }
            for (j = 0; j < 6; j++) {
                if (k < lo)
                    u16_write(dir->vol->sector_buffer, offset + 14 + j * 2, *(rf_u16_t*)(long_name + ((i - 1) * 26) + k));
                k += 2;
            }
            for (j = 0; j < 2; j++) {
                if (k < lo)
                    u16_write(dir->vol->sector_buffer, offset + 28 + j * 2, *(rf_u16_t*)(long_name + ((i - 1) * 26) + k));
                k += 2;
            }
        }
        lo = 26;
        offset += 32;
        if (offset == file->vol->sector_size) {
            ret = rf_disk_write(dir->vol->dev_handle, dir->vol->current_work_sector, 1, dir->vol->sector_buffer);
            if (ret)
                return ret;

            ret = fat_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &read_len);
            if (ret != RF_ESUCCESS)
                return ret;
            if (read_len == 0) {
                ret = fat_dir_add_clst(dir);
                if (ret != RF_ESUCCESS)
                    return ret;
            }
            offset = 0;
        }
    }

    file->entry.sector = dir->vol->current_work_sector;
    file->entry.offset = offset;
    return 0;
}

rf_int_t fat_file_set_long_name_imp(rf_fat_file_ref_t file, rf_u32_t sum, const rf_wchar_t* lfn)
{
    rf_int_t ret;
    rf_u32_t sector_offset;
    rf_u32_t file_offset, offset, read_len;
    rf_fat_file_t tdir;

    memset(&tdir, 0, sizeof(tdir));
    tdir.vol = file->vol;
    tdir.start_cluster = file->updir_start_cluster;
    ret = fat_dir_open(&tdir, RF_O_RDWR);
    if (ret != RF_ESUCCESS)
        return ret;

    file_offset = file->up_dir_offset;
    sector_offset = file_offset % file->vol->sector_size;
    ret = fat_file_lseek(&tdir, file_offset - sector_offset, RF_SEEK_SET);
    if (ret != RF_ESUCCESS)
        return ret;

    for (;;) {
        ret = fat_file_read(&tdir, file->vol->sector_buffer, file->vol->sector_size, &read_len);
        if (ret != RF_ESUCCESS)
            return ret;
        if (read_len == 0) {
            ret = fat_dir_add_clst(&tdir);
            if (ret != RF_ESUCCESS)
                return ret;

            /* ½Ĵ ʼֽΪ0 */
            offset = 0;
            goto _next;
        }

        for (offset = 0; offset < tdir.vol->sector_size; offset += 32)
            if (*(file->vol->sector_buffer + offset) == 0)
                goto _next;
    }

_next:
    return fat_file_write_long_name(&tdir, file, (rf_u32_t)offset, sum, (const rf_char_t*)lfn);
}

rf_int_t fat_file_set_long_name(rf_fat_file_ref_t file, const rf_wchar_t* lfn)
{
    rf_int_t ret;
    rf_u8_t entry[32];

    ret = fat_file_rm_long_name(file);
    if (RF_ESUCCESS != ret)
        return ret;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    memcpy(entry, file->vol->sector_buffer + file->entry.offset, 32);
    u8_write(file->vol->sector_buffer, file->entry.offset, FILE_DELETED);
    ret = rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    ret = fat_file_set_long_name_imp(file, fat_long_file_name_sum(entry), lfn);
    if (ret != RF_ESUCCESS)
        return ret;

    /* λдĿ¼ */
    memcpy(file->vol->sector_buffer + file->entry.offset, entry, 32);
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_GET_FSTAT
rf_int_t fat_file_get_stat(rf_fat_file_ref_t file, rf_stat_ref_t st)
{
    if (file->root_dir_flag) {
        memset(st, 0, sizeof(*st));
        st->st_mode = RF_ATTR_DIRECTORY;
        return RF_ESUCCESS;
    }
    st->entry.sector = file->entry.sector;
    st->entry.offset = file->entry.offset;
    st->current_offset = file->current_offset;
    st->st_mode = file->attribute;
    st->st_size = file->size;
    return RF_ESUCCESS;
}
#endif

#if RF_OPT_USE_GET_FTIME
rf_int_t fat_file_get_ftime(rf_fat_file_ref_t file, rf_file_ftime_ref_t ft)
{
    rf_int_t ret;
    /* ȡĿ¼ļʱ */
    if (file->root_dir_flag) {
        memset(&ft, 0, sizeof(ft));
        return RF_ESUCCESS;
    }

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    ft->create.low_date_time = u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtDate) << 16 |
                               u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime);
    ft->create.high_date_time = 0;
    ft->last_write.low_date_time = u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtDate) << 16 |
                                   u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime);
    ft->last_write.high_date_time = 0;
    /* Last access date. Note that there is no last access time, only a date.
    This is the date of last read or write. In the case of a write,
    this should be set to the same date as DIR_WrtDate. */
    ft->last_access.low_date_time = u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate) << 16;
    ft->last_access.high_date_time = 0;
    return RF_ESUCCESS;
}
#endif

#if RF_OPT_USE_GET_LTIME
static void fat_time_f2l(rf_u32_t ft, rf_tm_t* t)
{
    t->tm_year = ((ft >> 16) >> 9) + 1980;
    t->tm_mon = ((ft >> 16) >> 5) & 0xF;
    t->tm_mday = ((ft >> 16) & 0x1F);

    t->tm_hour = ((rf_u16_t)ft >> 11);
    t->tm_min = ((rf_u16_t)ft >> 5) & 0x3F;
    t->tm_sec = ((rf_u16_t)ft & 0x1F) << 1;
}

rf_int_t fat_file_get_ltime(rf_fat_file_ref_t file, rf_file_ltime_ref_t lt)
{
    rf_int_t ret;
    rf_u32_t t;

    /* ȡĿ¼ļʱ */
    if (file->root_dir_flag) {
        memset(&lt, 0, sizeof(lt));
        return RF_ESUCCESS;
    }
    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    t = u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime);
    fat_time_f2l(t, &lt->create);

    t = u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime);
    fat_time_f2l(t, &lt->last_write);

    /* Last access date. Note that there is no last access time, only a date.
    This is the date of last read or write. In the case of a write,
    this should be set to the same date as DIR_WrtDate. */
    t = u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate) << 16;
    fat_time_f2l(t, &lt->last_access);
    return RF_ESUCCESS;
}
#endif

rf_int_t fat_file_read(rf_fat_file_ref_t file, rf_void_t* buf, rf_u32_t len, rf_u32_t* plen)
{
    rf_int_t ret;
    rf_u32_t read, tmp, tmp1;

    *plen = 0;
    RF_PERROR("err: Can't read the file\n",
              ((file->attribute == RF_ATTR_LFN_ENTRY) || (file->attribute == RF_ATTR_VOLUME_ID)), return RF_EINVAL);
    RF_PERROR("err: access_flag == RF_O_WRONLY\n",
              (file->access_flag == RF_O_WRONLY), return RF_EACCES;);

    /* Ŀ¼ļ ûļȴС м ؽΪĿ¼ */
    if (!(file->attribute & RF_ATTR_DIRECTORY)) {
        tmp = file->size - file->current_offset;
        len = min(len, tmp);
    }
    if (len == 0)
        return RF_ESUCCESS;

    if (file->clust.byte_offset != 0) {
_READ_LAST:
        if (buf)
            ret = clusters_file_read(file, 1, &read, file->vol->sector_buffer);
        else
            ret = clusters_file_read(file, 1, &read, NULL);
        if (ret != RF_ESUCCESS)
            return ret;

        if (read == 1) {
            tmp = min((rf_u32_t)(file->vol->sector_size - file->clust.byte_offset), len);
            /* buf=pff->vol->sector_buffer ڶȡĿ¼ Ҫ */
            if (buf && buf != file->vol->sector_buffer)
                memcpy((rf_u8_t*)buf + *plen, file->vol->sector_buffer + file->clust.byte_offset, tmp);
            len -= tmp;
            file->clust.byte_offset += (rf_u16_t)tmp;
            if (file->clust.byte_offset % file->vol->sector_size == 0)
                file->clust.byte_offset = 0;
            *plen += tmp;
            file->current_offset += tmp;
        } else {
            /* ļôض ļȽС ֱӷ */
            return RF_ESUCCESS;
        }

        if (len == 0) {
            if (file->clust.byte_offset != 0)
                clusters_back_offset_sector(file, 1);
            return RF_ESUCCESS;
        }
    }

    read = len / file->vol->sector_size;
    if (read) {
        if (buf)
            ret = clusters_file_read(file, read, &tmp, (rf_u8_t*)buf + *plen);
        else
            ret = clusters_file_read(file, read, &tmp, NULL);
        if (ret != RF_ESUCCESS)
            return ret;

        tmp1 = tmp << file->vol->sector_size_bits;
        len -= tmp1;
        *plen += tmp1;
        file->current_offset += tmp1;
        /* ȡСҪʱļ */
        if (tmp < read)
            return RF_ESUCCESS;
    }

    /* ȡβһ */
    len %= file->vol->sector_size;
    if (len)
        goto _READ_LAST;
    return RF_ESUCCESS;
}

#if RF_OPT_USE_FAT12 || RF_OPT_USE_FAT16
static rf_int_t fat_dir_root_leek_12_16(rf_fat_file_ref_t file, rf_int_t offset, rf_int_t origin)
{
    switch (origin) {
    case RF_SEEK_CUR:
    case RF_SEEK_END:
        return RF_EPERM;

    case RF_SEEK_SET:
        /* file->length ƫ */
        file->size = offset << file->vol->sector_size_bits;
        file->current_offset = offset;
        return RF_ESUCCESS;
    }
    return RF_EINVAL;
}
#endif

rf_int_t fat_file_lseek(rf_fat_file_ref_t file, rf_u32_t offset, rf_int_t origin)
{
    rf_int_t ret;
    rf_u32_t tmp_offset;

    /* Ӽǲ fat12 16 Ŀ¼ */
#if RF_OPT_USE_FAT12 || RF_OPT_USE_FAT16
    if (file->root_dir_flag && (file->vol->fs_type == RF_TYPE_FAT12 || file->vol->fs_type == RF_TYPE_FAT16))
        return fat_dir_root_leek_12_16(file, (rf_int_t)offset, origin);
#endif

    switch (origin) {
    /* ļǰλÿʼƫ */
    case RF_SEEK_CUR:
        ret = fat_file_read(file, NULL, (rf_u32_t)offset, &tmp_offset);
        if (ret)
            return ret;
        file->current_offset += tmp_offset;
        break;

    /* ƫƵļʼ */
    case RF_SEEK_SET:
        file->first_gather_cluester_bit = FALSE;
        ret = clusters_gather_file(file, TRUE);
        if (ret)
            return ret;
        file->current_offset = 0;
        ret = fat_file_read(file, NULL, (rf_u32_t)offset, &tmp_offset);
        if (ret)
            return ret;
        file->current_offset = tmp_offset;
        break;

    /* ƫƵļλ */
    case RF_SEEK_END:
        if (offset != 0)
            return RF_EINVAL;
        tmp_offset = file->size - file->current_offset;
        ret = fat_file_read(file, NULL, (rf_u32_t)tmp_offset, &tmp_offset);
        if (ret)
            return ret;
        file->current_offset = file->size;
        break;

    default:
        return RF_EINVAL;
    }
    return RF_ESUCCESS;
}

#if RF_OPT_FS_READONLY == 0

#if RF_OPT_USE_RENAME
rf_int_t fat_file_set_stat(rf_fat_file_ref_t file, rf_file_ftime_ref_t st)
{
    rf_int_t ret;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    u8_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_Attr, (rf_u8_t)file->attribute);
    u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime, st->create.low_date_time);
    u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime, st->last_write.low_date_time);
    u16_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate, (rf_u16_t)(st->last_access.low_date_time >> 16));

    /* ļʼ */
    u16_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 20), (rf_u16_t)(file->start_cluster >> 16));
    u16_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 26), (rf_u16_t)(file->start_cluster));
    /* ļ */
    u32_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 28), file->size);
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_SET_ATTRIB
rf_int_t fat_file_set_attribute(rf_fat_file_ref_t file, rf_u32_t attribute)
{
    rf_int_t ret;
    rf_u32_t tmp, bwrite = FALSE;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    tmp = u8_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_Attr);
    if (attribute != tmp) {
        RF_PERROR("err: Can't change the file or directory attribute\n",
                  ((attribute & RF_ATTR_DIRECTORY) != (tmp & RF_ATTR_DIRECTORY)), return RF_EINVAL;);
        bwrite = TRUE;
        u8_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_Attr, (rf_u8_t)attribute);
    }
    file->attribute = attribute;

    if (!bwrite)
        return 0;
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_SET_FTIME
rf_int_t fat_file_set_ftime(rf_fat_file_ref_t file, rf_file_ftime_ref_t st)
{
    rf_int_t ret;
    rf_u32_t bwrite = FALSE;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    if (st->create.low_date_time != u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime)) {
        bwrite = TRUE;
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime, st->create.low_date_time);
    }

    if (st->last_write.low_date_time != u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime)) {
        bwrite = TRUE;
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime, st->last_write.low_date_time);
    }

    if ((rf_u16_t)(st->last_access.low_date_time >> 16) != u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate)) {
        bwrite = TRUE;
        u16_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate, (rf_u16_t)(st->last_access.low_date_time >> 16));
    }

    if (!bwrite)
        return 0;
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_SET_LTIME
rf_int_t fat_file_set_ltime(rf_fat_file_ref_t file, rf_file_ltime_ref_t lt)
{
    rf_int_t ret;
    rf_u32_t tmp;
    rf_u32_t bwrite = FALSE;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    tmp = TM2FT(lt->create);
    if (tmp != u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime)) {
        bwrite = TRUE;
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime, tmp);
    }

    tmp = TM2FT(lt->last_write);
    if (tmp != u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime)) {
        bwrite = TRUE;
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime, tmp);
    }

    tmp = TM2FT(lt->last_access) >> 16;
    if ((rf_u16_t)tmp != u16_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate)) {
        bwrite = TRUE;
        u16_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate, (rf_u16_t)tmp);
    }

    if (!bwrite)
        return 0;
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_SET_NAME
rf_int_t fat_file_set_name(rf_fat_file_ref_t file, const rf_char_t* name)
{
    rf_int_t ret;
    rf_char_t tbuf[12];
    rf_fat_file_t tdir;

    if (strlen(name) > 12)
        return RF_EINVAL;
    ret = fat_convert_short_name(name, tbuf);
    if (ret)
        return ret;

    memset(&tdir, 0, sizeof(tdir));
    tdir.vol = file->vol;
    tdir.start_cluster = file->updir_start_cluster;
    ret = fat_dir_open(&tdir, RF_O_RDONLY);
    if (ret != RF_ESUCCESS)
        return ret;

    ret = fat_file_lseek(&tdir, file->up_dir_offset, RF_SEEK_SET);
    if (ret != RF_ESUCCESS)
        return ret;

    /* һĿ¼Ƿ ͬļ */
    ret = fat_file_read_dir(&tdir, name, NULL, RF_O_RAW_SFN);
    if (ret == RF_ESUCCESS)
        return RF_ENOENT;
    if (ret != RF_ENDFILE)
        return ret;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    memcpy(file->vol->sector_buffer + file->entry.offset, tbuf, 11);
    u8_write((file->vol->sector_buffer + file->entry.offset), DIR_NTres, tbuf[11]);
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

static rf_int_t fat_file_update_time(rf_fat_file_ref_t file, rf_int_t flags)
{
    rf_int_t ret;
    rf_u32_t t = fat_time_current();

    /* Ŀ¼Ҫʱ */
    if (file->root_dir_flag)
        return 0;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    if (flags == RF_FILE_CREATE_FLAG) {
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_CrtTime, t);
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime, t);
        u16_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_LstAccDate, 0);
    }
    if (flags & RF_FILE_WRITE_FLAG) {
        u32_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + DIR_WrtTime, t);
    }
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}

rf_int_t fat_file_sync(rf_fat_file_ref_t file)
{
    rf_int_t ret;

    if (file->open_flag && file->attrib_change_bit) {
        file->attrib_change_bit = FALSE;
        ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
        if (ret)
            return ret;

        /* дʱ */
        u32_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + DIR_WrtTime), fat_time_current());
        u16_write(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + 18, 0);
        /* ļʼ */
        u16_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 20), (rf_u16_t)(file->start_cluster >> 16));
        u16_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 26), (rf_u16_t)(file->start_cluster));
        /* ļ */
        u32_write(file->vol->sector_buffer, (rf_u16_t)(file->entry.offset + 28), file->size);

        ret = rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
        if (ret)
            return ret;
    }
    return RF_ESUCCESS;
}

#define FAT_MAX_FILE_SIZE 0xffffffff
rf_int_t fat_file_write(rf_fat_file_ref_t file, rf_void_t* buf, rf_u32_t len, rf_u32_t* plen)
{
    rf_int_t ret;
    rf_u32_t tmp, count, tmp1;

    if (!(file->access_flag & 0x3) || file->attribute & RF_ATTR_DIRECTORY)
        return RF_EACCES;
    if (!len)
        return RF_ESUCCESS;
    if (file->size == FAT_MAX_FILE_SIZE)
        return RF_EFBIG;

    /* fatļϵͳ ļ4g=0xffffffff */
    len = min(len, FAT_MAX_FILE_SIZE - file->size);
    *plen = 0;
    if (file->clust.byte_offset != 0) {
_WRITE_LAST:
        if (buf)
            ret = clusters_file_read(file, 1, &count, file->vol->sector_buffer);
        else
            ret = clusters_file_read(file, 1, &count, NULL);
        if (ret != RF_ESUCCESS)
            return ret;
        /* ½ ʼһ */
        if (count == 0) {
            tmp = len;
            ret = clusters_file_write_sector(file, NULL, 1, &count);
            if (ret)
                return ret;
            /* ֮ ƫƶfatʱ ƻ*/
            if (buf) {
                memcpy(file->vol->sector_buffer, (rf_u8_t*)buf + *plen, len);
                memset(file->vol->sector_buffer + len, 0, file->vol->sector_size - len);
                clusters_back_offset_sector(file, 1);
                ret = clusters_file_write_sector(file, file->vol->sector_buffer, 1, &count);
                if (ret)
                    return ret;
            }
        } else {
            tmp = min((rf_u32_t)(file->vol->sector_size - file->clust.byte_offset), len);
            if (buf) {
                memcpy(file->vol->sector_buffer + file->clust.byte_offset, (rf_u8_t*)buf + *plen, tmp);
                clusters_back_offset_sector(file, 1);
                ret = clusters_file_write_sector(file, file->vol->sector_buffer, 1, &count);
                if (ret)
                    return ret;
            }
        }

        len -= tmp;
        file->clust.byte_offset += (rf_u16_t)tmp;
        if (file->clust.byte_offset % file->vol->sector_size == 0)
            file->clust.byte_offset = 0;
        else
            clusters_back_offset_sector(file, 1);
        *plen += tmp;
        file->current_offset += tmp;
        if (!len)
            goto _EXIT;
    }

    /* ʼд */
    tmp = len >> file->vol->sector_size_bits;
    if (tmp) {
        if (buf)
            ret = clusters_file_write_sector(file, (rf_u8_t*)buf + *plen, tmp, &count);
        else
            ret = clusters_file_write_sector(file, NULL, tmp, &count);
        if (ret)
            return ret;

        tmp1 = count << file->vol->sector_size_bits;
        file->current_offset += tmp1;
        *plen += tmp1;
        len -= tmp1;
        /* ûпÿռ ֱӷ */
        if (count < tmp)
            return RF_ESUCCESS;
    }

    len %= file->vol->sector_size;
    if (len)
        goto _WRITE_LAST;
_EXIT:
    if (file->current_offset > file->size)
        file->size = file->current_offset;
    /* ļ ܸı ļʱ һı */
    file->attrib_change_bit = TRUE;
    return RF_ESUCCESS;
}

#if RF_OPT_USE_FTRUNCATE
rf_int_t fat_file_truncate(rf_fat_file_ref_t file, rf_u32_t length)
{
    rf_int_t ret;
    rf_u32_t len;
    rf_u8_t attr;

    if (!file->access_flag || file->attribute & RF_ATTR_DIRECTORY)
        return RF_EACCES;

    /* ܲüļ ɾĿ¼ */
    RF_PERROR("err: Can't truncate the file\n",
              ((file->attribute == RF_ATTR_LFN_ENTRY) || (file->root_dir_flag)), return RF_EINVAL;);

    if (length == file->size)
        return RF_ESUCCESS;
    if (length < file->size) {
        /* ƫļָ üƫƳ */
        len = align_size_u32(length, file->vol->sector_size << file->vol->sectors_per_cluster_bits);
        /* ʱıļΪĿ¼ ƫƵؽβʱ ļȼ */
        attr = file->attribute;
        file->attribute |= RF_ATTR_DIRECTORY;
        ret = fat_file_lseek(file, len, RF_SEEK_SET);
        file->attribute = attr;
        if (ret)
            return ret;

        /* ɾǰƫƺĴ */
        ret = clusters_file_delete(file);
        if (RF_ESUCCESS != ret)
            return ret;
    } else {
        /* ƫƵļβ */
        ret = fat_file_lseek(file, 0, RF_SEEK_END);
        if (ret)
            return ret;
        /* ܿռ䲻 */
        ret = fat_file_write(file, NULL, length - file->size, &len);
        if (ret)
            return ret;
        length = file->size;
    }

    file->attrib_change_bit = TRUE;
    file->size = length;
    if (length == 0)
        file->start_cluster = 0;
    return fat_file_lseek(file, file->current_offset, RF_SEEK_SET);
}
#endif

#if RF_OPT_USE_FDELETE
static rf_int_t fat_file_del_clust(rf_fat_file_ref_t file)
{
    rf_int_t ret;

    /* ɾļ ɾĿ¼ */
    RF_PERROR("err: can't delete the file\n",
              ((file->attribute == RF_ATTR_LFN_ENTRY) || file->root_dir_flag), return RF_EACCES;);

    /* ļд */
    if (file->start_cluster != 0) {
        file->first_gather_cluester_bit = FALSE;
        ret = clusters_gather_file(file, TRUE);
        if (RF_ESUCCESS != ret)
            return ret;

        /* մ */
        file->current_offset = 0;
        ret = clusters_file_delete(file);
        if (RF_ESUCCESS != ret)
            return ret;
    }
    return 0;
}

rf_int_t fat_file_delete(rf_fat_file_ref_t file)
{
    rf_int_t ret;

    ret = fat_file_del_clust(file);
    if (ret)
        return ret;

#if RF_OPT_FAT_LONG_NAME
    /* ճļ */
    ret = fat_file_rm_long_name(file);
    if (RF_ESUCCESS != ret)
        return ret;
#endif

    /* Ŀ¼ */
    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;
    u8_write(file->vol->sector_buffer, file->entry.offset, FILE_DELETED);
    return rf_disk_write(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
}
#endif

#if RF_OPT_USE_RENAME
/* ǲǸĿ¼ϵ 0 ΪΪ */
static rf_int_t fat_dir_check_sub(rf_fat_file_ref_t parent, rf_fat_file_ref_t son)
{
    rf_int_t  ret;
    rf_dirent_t dir;
    rf_fentry_t entry;
    rf_char_t buf[12];

    entry = son->entry;
    strcpy(buf, "..         ");
    for (;;) {
        /* ζȡϼĿ¼ .. ǲ Ŀ¼ڴ */
        dir.name = NULL;
        ret = fat_file_read_dir(son, buf, &dir, RF_O_RAW_SFN);
        if (ret)
            return ret;

        ret = fat_file_raw_open(son, &dir.entry, RF_O_RDONLY);
        if (ret)
            return ret;
        RF_PERROR("err: rename can't to the parent directory\n",
                  (parent->start_cluster == son->start_cluster), return RF_EINVAL;);
        if (son->root_dir_flag)
            break;
    }
    return 0;
}

/* Ŀ¼ . ..  ƶĿ¼ */
static rf_int_t fat_dir_dot_copy(rf_fat_file_ref_t dest, rf_fat_file_ref_t src)
{
    rf_int_t ret;
    rf_u32_t len, now_clust, up_clust;
    rf_u8_t* buf;

    buf = src->vol->sector_buffer;
    ret = clusters_file_read(src, 1, &len, buf);
    if (ret)
        return ret;

    now_clust = u16_read(buf, DIR_FstClusHI) << 16 | u16_read(buf, DIR_FstClusLO);
    up_clust = u16_read(buf + 32, DIR_FstClusHI) << 16 | u16_read(buf + 32, DIR_FstClusLO);

    buf = src->vol->sector_buffer;
    ret = clusters_file_read(src, 1, &len, buf);
    if (ret)
        return ret;

    if (now_clust == (u16_read(buf, DIR_FstClusHI) << 16 | u16_read(buf, DIR_FstClusLO))
            && up_clust == (u16_read(buf + 32, DIR_FstClusHI) << 16 | u16_read(buf + 32, DIR_FstClusLO))) {
        return 0;
    }

    /* ļʼ */
    u16_write(buf, DIR_FstClusHI, (rf_u16_t)(now_clust >> 16));
    u16_write(buf, DIR_FstClusLO, (rf_u16_t)(now_clust));
    u16_write(buf + 32, DIR_FstClusHI, (rf_u16_t)(up_clust >> 16));
    u16_write(buf + 32, DIR_FstClusLO, (rf_u16_t)(up_clust));

    clusters_back_offset_sector(dest, 1);
    return clusters_file_write_sector(dest, buf, 1, &len);
}

rf_int_t fat_file_rename(rf_fat_file_ref_t oldfile, rf_fat_file_ref_t newfile, const rf_char_t* oldname, const rf_char_t* newname)
{
    rf_int_t ret;
    rf_int_t flag;
    rf_fentry_t entry;
    rf_file_ftime_t time;

    /* ļ дȨ */
    ret = fat_file_open(oldfile, oldname, RF_O_RDWR);
    if (ret)
        return ret;

    /* ȡ   */
    ret = fat_file_get_ftime(oldfile, &time);
    if (ret)
        return ret;

    /* ļ ļü */
    flag = RF_O_RDWR | RF_O_CREAT | RF_O_TRUNC;
    if (oldfile->attribute & RF_ATTR_DIRECTORY)
        flag |= RF_O_DIRECTORY;
    ret = fat_file_open(newfile, newname, flag);
    if (ret)
        return ret;

    /* Ŀ¼ǲ Ŀ¼Ŀ¼ */
    if (oldfile->attribute & RF_ATTR_DIRECTORY) {
        entry = newfile->entry;
        ret = fat_dir_check_sub(oldfile, newfile);
        if (ret)
            return ret;
        /* ´Ŀ¼ */
        ret = fat_file_raw_open(newfile, &entry, RF_O_RDWR);
        if (ret)
            return ret;
    }

    if (oldfile->attribute & RF_ATTR_DIRECTORY) {
        /* ļĿ¼ϼغ͵ǰ  Ŀ¼ */
        ret = fat_dir_dot_copy(oldfile, newfile);
        if (ret)
            return ret;

        /* ɾĿ¼  */
        ret = fat_file_del_clust(newfile);
        if (ret)
            return ret;

        /* ļ ļһ */
        newfile->attribute = oldfile->attribute;
        newfile->start_cluster = oldfile->start_cluster;
        ret = fat_file_set_stat(newfile, &time);
    } else {
        /* ļ ļһ */
        newfile->attribute = oldfile->attribute;
        newfile->start_cluster = oldfile->start_cluster;
        newfile->size = oldfile->size;
        ret = fat_file_set_stat(newfile, &time);
        if (ret)
            return ret;
    }

    /* ɾļ ûɾ */
    oldfile->start_cluster = 0;
    return fat_file_delete(oldfile);
}
#endif

#endif
