/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "fat_volume.h"
#include "fat_file.h"

#define BS_jmpBoot          0       /* x86 jump instruction (3-byte) */
#define BS_OEMName          3       /* OEM name (8-byte) */
#define BPB_BytsPerSec      11      /* Sector size [byte] (WORD) */
#define BPB_SecPerClus      13      /* Cluster size [sector] (BYTE) */
#define BPB_RsvdSecCnt      14      /* Size of reserved area [sector] (WORD) */
#define BPB_NumFATs         16      /* Number of FATs (BYTE) */
#define BPB_RootEntCnt      17      /* Size of root directory area for FAT12/16 [entry] (WORD) */
#define BPB_TotSec16        19      /* Volume size (16-bit) [sector] (WORD) */
#define BPB_Media           21      /* Media descriptor (BYTE) */
#define BPB_FATSz16         22      /* FAT size (16-bit) [sector] (WORD) */
#define BPB_SecPerTrk       24      /* Track size for int13h [sector] (WORD) */
#define BPB_NumHeads        26      /* Number of heads for int13h (WORD) */
#define BPB_HiddSec         28      /* Volume offset from top of the drive (DWORD) */
#define BPB_TotSec32        32      /* Volume size (32-bit) [sector] (DWORD) */
#define BS_DrvNum           36      /* Physical drive number for int13h (BYTE) */
#define BS_NTres            37      /* Error flag (BYTE) */
#define BS_BootSig          38      /* Extended boot signature (BYTE) */
#define BS_VolID            39      /* Volume serial number (DWORD) */
#define BS_VolLab           43      /* Volume label string (8-byte) */
#define BS_FilSysType       54      /* File system type string (8-byte) */
#define BPB_FATSz32         36      /* FAT size (32-bit) [sector] (DWORD) */
#define BPB_ExtFlags        40      /* Extended flags (WORD) */

#define BPB_FSVer32         42      /* FAT32: File system version (WORD) */
#define BPB_RootClus32      44      /* FAT32: Root directory cluster (DWORD) */
#define BPB_FSInfo32        48      /* FAT32: Offset of FSINFO sector (WORD) */
#define BPB_BkBootSec32     50      /* FAT32: Offset of backup boot sector (WORD) */
#define BS_DrvNum32         64      /* FAT32: Physical drive number for int13h (BYTE) */
#define BS_NTres32          65      /* FAT32: Error flag (BYTE) */
#define BS_BootSig32        66      /* FAT32: Extended boot signature (BYTE) */
#define BS_VolID32          67      /* FAT32: Volume serial number (DWORD) */
#define BS_VolLab32         71      /* FAT32: Volume label string (8-byte) */
#define BS_FilSysType32     82      /* FAT32: File system type string (8-byte) */

#define BPB_ZeroedEx        11      /* exFAT: Must be zero (35-byte) */
#define BPB_VolOfsEx        64      /* exFAT: Volume offset from top of the drive [sector] (QWORD) */
#define BPB_TotSecEx        72      /* exFAT: Volume size [sector] (QWORD) */
#define BPB_FatOfsEx        80      /* exFAT: FAT offset from top of the volume [sector] (DWORD) */
#define BPB_FatSzEx         84      /* exFAT: FAT size [sector] (DWORD) */
#define BPB_DataOfsEx       88      /* exFAT: Data offset from top of the volume [sector] (DWORD) */
#define BPB_NumClusEx       92      /* exFAT: Number of clusters (DWORD) */
#define BPB_RootClusEx      96      /* exFAT: Root directory cluster (DWORD) */
#define BPB_VolIDEx         100     /* exFAT: Volume serial number (DWORD) */
#define BPB_FSVerEx         104     /* exFAT: File system version (WORD) */
#define BPB_VolFlagEx       106     /* exFAT: Volume flags (BYTE) */
#define BPB_ActFatEx        107     /* exFAT: Active FAT flags (BYTE) */
#define BPB_BytsPerSecEx    108     /* exFAT: Log2 of sector size in byte (BYTE) */
#define BPB_SecPerClusEx    109     /* exFAT: Log2 of cluster size in sector (BYTE) */
#define BPB_NumFATsEx       110     /* exFAT: Number of FATs (BYTE) */
#define BPB_DrvNumEx        111     /* exFAT: Physical drive number for int13h (BYTE) */
#define BPB_PercInUseEx     112     /* exFAT: Percent in use (BYTE) */

rf_int_t clusters_gather_empty(rf_fat_file_ref_t file, rf_u32_t number_clst, rf_u32_ref_t pnumber_clst, rf_bool_t gather);

#if RF_OPT_USE_EXFAT
static rf_int_t exfat_init_volume(rf_fat_volume_ref_t vol)
{
    rf_int_t i;
    rf_u64_t maxlba;

    for (i = BPB_ZeroedEx; i < BPB_ZeroedEx + 53 && !vol->sector_buffer[i]; i++);   /* Check zero filler */
    if (i < BPB_ZeroedEx + 53)
        return RF_ENOFS;
    /* Check exFAT revision (Must be 1.0) */
    if (u32_read(vol->sector_buffer, BPB_FSVerEx) != 0x100)
        return RF_ENOFS;

    if (1 << vol->sector_buffer[BPB_BytsPerSecEx] != vol->sector_size)
        return RF_ENOFS;

    maxlba = u64_read(vol->sector_buffer, BPB_TotSecEx);
    if (maxlba >= 0x100000000)
        return RF_ENOFS;

    vol->bpb_fat_size = u32_read(vol->sector_buffer, BPB_FatSzEx);
    vol->bpb_num_fats = vol->sector_buffer[BPB_NumFATsEx];
    if (vol->bpb_num_fats != 1)
        return RF_ENOFS;    /* (Must be 1) */

    vol->sectors_per_cluster = 1 << vol->sector_buffer[BPB_SecPerClusEx];
    if (vol->sectors_per_cluster == 0)
        return RF_ENOFS;    /* (Must be 1..32768) */

    vol->bpb_total_clusters = u32_read(vol->sector_buffer, BPB_NumClusEx);
    if (vol->bpb_total_clusters + 2 >= 0x80000000)
        return RF_ENOFS;    /* (Must be <= 0x7FFFFFFF) */

    vol->bpb_fat_start_sector = u32_read(vol->sector_buffer, BPB_FatOfsEx);
    vol->fat_data_start_sector = u32_read(vol->sector_buffer, BPB_DataOfsEx);
    if (maxlba < vol->fat_data_start_sector + vol->bpb_total_clusters * vol->sectors_per_cluster)
        return RF_ENOFS;    /* (Volume size must not be smaller than the size requiered) */
    vol->bpb_root_start_sector = u32_read(vol->sector_buffer, BPB_RootClusEx);

#if 0
    /* Check if bitmap location is in assumption (at the first cluster) */
    if (move_window(fs, clust2sect(fs, fs->dirbase)) != FR_OK)
        return FR_DISK_ERR;
    for (i = 0; i < SS(fs); i += SZDIRE) {
        if (fs->win[i] == 0x81 && ld_dword(fs->win + i + 20) == 2)
            break;
    }
    if (i == SS(fs))
        return FR_NO_FILESYSTEM;
    fs->last_clst = fs->free_clst = 0xFFFFFFFF;     /* Initialize cluster allocation information */
#endif

    return 0;
}
#endif

rf_int_t fat_volume_open(rf_fat_volume_ref_t vol, rf_mount_param_t* param)
{
    rf_int_t ret;
    rf_u32_t sector_bytes, total_sector, sys_sector, fat_size, fat_need_sector;
    rf_u16_t rsvd_sector, sectors_per_fat_16;

    ret = rf_disk_open(param->device_name, param->flags, param->ctx, &vol->dev_handle);
    if (ret)
        return ret;
    vol->access_flag = 0;
    if ((param->flags & 0x3) == 2)
        vol->access_flag = 1;

    sector_bytes = param->ctx->sector_size;
    if (sector_bytes == 0)
        ret = rf_disk_ioctl(vol->dev_handle, RF_CMD_GET_SECTOR_SIZE, &sector_bytes, sizeof(sector_bytes));
    if (ret)
        return ret;
    RF_PERROR("sector bytes error", INVOID_SECTOR_SIZE(sector_bytes), return RF_ENOFS);

#if RF_OPT_USE_MALLOC
    vol->sector_buffer = malloc(sector_bytes);
    if (vol->sector_buffer == NULL)
        return RF_ENOMEM;
#else
    if (RF_OPT_MAX_SECTOR_SIZE < sector_bytes)
        return RF_EBUFSMALL;
#endif
    /* ڲļ豸 ĬСΪ512ֽ */
    vol->sector_size = (rf_u16_t)sector_bytes;
    vol->sector_size_bits = rf_ffs(vol->sector_size);

    ret = rf_disk_read(vol->dev_handle, 0, 1, vol->sector_buffer);
    if (ret)
        return ret;
    if (u16_read(vol->sector_buffer, 510) != 0xAA55)
        return RF_ENOFS;

#if RF_OPT_USE_EXFAT
    if (!memcmp(vol->sector_buffer + BS_OEMName, "EXFAT   ", 8)) {
        vol->fs_type = RF_TYPE_EXFAT;
        return exfat_init_volume(vol);
    }
#endif

    /* һfatռ */
    fat_size = (rf_u32_t)u16_read(vol->sector_buffer, 22);
    sectors_per_fat_16 = fat_size;
    if (!fat_size)
        fat_size = u32_read(vol->sector_buffer, 36);
    vol->bpb_fat_size = fat_size;

    /* fat */
    vol->bpb_num_fats = vol->sector_buffer[16];
    if (vol->bpb_num_fats != 1 && vol->bpb_num_fats != 2)
        return RF_ENOFS;
    fat_size *= vol->bpb_num_fats;

    /* һؼ!! */
    vol->sectors_per_cluster = vol->sector_buffer[13];
    if (!vol->sectors_per_cluster || (vol->sectors_per_cluster & (vol->sectors_per_cluster - 1))) {
        RF_PRINTF(RF_DBG_LEVEL_ERROR, ("SectorsPerCluster %d\n", vol->sectors_per_cluster));
        return RF_ENOFS;
    }
    vol->sectors_per_cluster_bits = rf_ffs(vol->sectors_per_cluster);

    /* fat12/16 Ŀ¼ fat32Ϊ 0 */
    vol->bpb_root_sector_size = u16_read(vol->sector_buffer, 17);
    /*  */
    if (vol->bpb_root_sector_size % (vol->sector_size / 32))
        return RF_ENOFS;
    vol->bpb_root_sector_size /= (vol->sector_size / 32);

    /*  */
    total_sector = (rf_u32_t)u16_read(vol->sector_buffer, 19);
    if (!total_sector)
        total_sector = u32_read(vol->sector_buffer, 32);

    /*  */
    rsvd_sector = u16_read(vol->sector_buffer, 14);
    if (!rsvd_sector)
        return RF_ENOFS;

    sys_sector = rsvd_sector + fat_size + vol->bpb_root_sector_size;
    if (total_sector < sys_sector)
        return RF_ENOFS;

    /* ж */
    vol->bpb_total_clusters = (total_sector - sys_sector) / vol->sectors_per_cluster;
    if (vol->bpb_total_clusters == 0x0)
        return RF_ENOFS;

    if (sectors_per_fat_16 == 0) {
        vol->fs_type = RF_TYPE_FAT32;
#if !RF_OPT_USE_FAT32
        RF_PRINTF(RF_DBG_LEVEL_WARNING, ("error: RF_OPT_USE_FAT32==0\n"));
        return RF_ENOFS;
#endif
    } else {
        if (vol->bpb_total_clusters <= 4085 + 2) {
            vol->fs_type = RF_TYPE_FAT12;
#if !RF_OPT_USE_FAT12
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("error: RF_OPT_USE_FAT12==0\n"));
            return RF_ENOFS;
#endif
        } else {
            vol->fs_type = RF_TYPE_FAT16;
#if !RF_OPT_USE_FAT16
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("error: RF_OPT_USE_FAT16==0\n"));
            return RF_ENOFS;
#endif
        }
    }

    /* һfatʼַ */
    vol->bpb_fat_start_sector = rsvd_sector;
    vol->fat_data_start_sector = sys_sector;

    vol->fat_fsi_free_count = 0x0;
    vol->fat_fsi_nxt_free = 0x2;
    if (vol->fs_type == RF_TYPE_FAT32) {
#if RF_OPT_USE_FAT32
        vol->bpb_root_start_sector = u32_read(vol->sector_buffer, 44);
        fat_need_sector = (vol->bpb_total_clusters + 2) * 4;

#if RF_OPT_FS_READONLY == 0
        vol->fat_fsinfo_sector = u16_read(vol->sector_buffer, 48);
        ret = rf_disk_read(vol->dev_handle, vol->fat_fsinfo_sector, 1, vol->sector_buffer);
        if (ret)
            return ret;

#ifdef RF_DEBUG
        if (u32_read(vol->sector_buffer, 0) != 0x41615252)
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("FSI_LeadSig Error\n"));
        if (u32_read(vol->sector_buffer, 484) != 0x61417272)
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("FSI_StrucSig Error\n"));
        if (u32_read(vol->sector_buffer, 508) != 0xAA550000)
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("FSI_TrailSig Error\n"));
#endif
        /* fsiЧ!!! */
        vol->fat_fsi_free_count = u32_read(vol->sector_buffer, 488);
        vol->fat_fsi_nxt_free = u32_read(vol->sector_buffer, 492);
#endif
#endif
    } else {
#if RF_OPT_USE_FAT12 || RF_OPT_USE_FAT16
        vol->bpb_root_start_sector = vol->bpb_fat_start_sector + fat_size;
        fat_need_sector = (vol->fs_type == RF_TYPE_FAT16) ?
                          (vol->bpb_total_clusters + 2) * 2 : ((vol->bpb_total_clusters + 2) * 3) / 2 + ((vol->bpb_total_clusters + 2) & 1);
#endif
    }
    RF_PERROR("fs error\n", (vol->bpb_fat_size < (fat_need_sector + (vol->sector_size - 1)) / vol->sector_size), return RF_ENOFS);

    vol->current_work_sector = 0xFFFFFFFF;
    /* ʼǰĿ¼ ΪĿ¼ */
    vol->current_dir_entry.offset = 0;
    vol->current_dir_entry.sector = 0;
    return RF_ESUCCESS;
}

rf_int_t fat_volume_close(rf_fat_volume_ref_t vol)
{
    rf_int_t ret;

#if RF_OPT_FS_READONLY == 0
#if RF_OPT_USE_FAT32
    if (vol->fs_type == RF_TYPE_FAT32 && vol->update_fsi_flag) {
        ret = rf_disk_read(vol->dev_handle, vol->fat_fsinfo_sector, 1, vol->sector_buffer);
        if (ret)
            return ret;
        u32_write(vol->sector_buffer, 488, vol->fat_fsi_free_count);
        u32_write(vol->sector_buffer, 492, vol->fat_fsi_nxt_free);
        ret = rf_disk_write(vol->dev_handle, vol->fat_fsinfo_sector, 1, vol->sector_buffer);
        if (ret)
            return ret;
    }
#endif
#endif

    ret = rf_disk_close(vol->dev_handle);
    if (ret == RF_ESUCCESS) {
#if RF_OPT_USE_MALLOC
        if (vol->sector_buffer != NULL)
            rf_free(vol->sector_buffer);
#endif
    }
    return ret;
}

#if RF_OPT_USE_STATFS
rf_int_t fat_volume_stat(rf_fat_volume_ref_t vol, rf_statfs_ref_t st)
{
    rf_fat_file_t ff;
    rf_u32_t numclst;
    rf_int_t ret;

    vol->current_work_sector = 0xffffffff;
    st->f_type = vol->fs_type;
    st->f_ssize = vol->sector_size;
    st->f_bsize = vol->sectors_per_cluster << vol->sectors_per_cluster_bits;
    st->f_blocks = vol->bpb_total_clusters;

    ff.vol = vol;
    ret = clusters_gather_empty(&ff, 0xffffffff, &numclst, FALSE);
    if (ret != RF_ESUCCESS)
        return ret;
    st->f_bfree = numclst;
    return 0;
}
#endif

#if RF_OPT_USE_GET_LABEL
rf_int_t fat_volume_get_label(rf_fat_volume_ref_t vol, const rf_char_t* buf, rf_int_t len)
{
    return RF_EPERM;
}
#endif

#if RF_OPT_USE_CHDIR
rf_int_t fat_volume_set_current_dir(rf_fat_volume_ref_t vol, rf_fentry_t* entry)
{
    vol->current_dir_entry.sector = entry->sector;
    vol->current_dir_entry.offset = entry->offset;
    return 0;
}
#endif

#if RF_OPT_FS_READONLY == 0
#if RF_OPT_USE_SET_LABEL
rf_int_t fat_volume_set_label(rf_fat_volume_ref_t vol, const rf_char_t* buf)
{
    return RF_EPERM;
}
#endif
#endif
