/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "ntfs_volume.h"
#include "ntfs_file.h"
#include "ntfs.h"

/* ԺŻ*/
rf_u64_t mft2sec(rf_ntfs_volume_ref_t vol, rf_u64_t mft)
{
    rf_u64_t tmp, mft1, f1, f2;

    mft &= 0x0000ffffffffffff;
    if (vol->mft_sizeof_sector == 2)
        mft <<= 1;

    mft1 = mft;
    f1 = vol->mft_size[0] << vol->sectors_per_cluster_bits;       /* һMFTС */
    if (f1 > mft) {
        tmp = (vol->mft_lcn << vol->sectors_per_cluster_bits) + mft;
    } else {
        mft1 -= f1;
        f2 = vol->mft_size[1] << vol->sectors_per_cluster_bits;
        if (f2 > mft1) {
            tmp = (vol->mft_offet[1] << vol->sectors_per_cluster_bits) + mft - f1;
        } else {
            tmp = (vol->mft_offet[2] << vol->sectors_per_cluster_bits) + mft - f1 - f2;
        }
    }
    return tmp;
}

rf_s32_t get_s32(rf_u8_t* buf, rf_u16_t i, rf_u8_t cb)
{
    rf_u32_t d;

    if (!cb)
        return 0;
    d = u32_read(buf, i);
    switch (cb) {
    case 1:
        d &= 0xff;
        /* жd ǷǸ */
        if (d & 0x80)
            d |= 0xffffff00;
        return d;

    case 2:
        d &= 0xffff;
        if (d & 0x8000)
            d |= 0xffff0000;
        return d;

    case 3:
        d &= 0xffffff;
        if (d & 0x800000)
            d |= 0xff000000;
        return d;

    default:
        return (rf_s32_t)d;
    }
}

/* All important structures in NTFS use 2 consistency checks:
 * . a magic structure identifier (FILE, INDX, RSTR, RCRD...)
 * . a fixup technique : the last word of each sector (called a fixup) of a
 *   structure's record should end with the word at offset <n> of the first
 *   sector, and if it is the case, must be replaced with the words following
 *   <n>. The value of <n> and the number of fixups is taken from the fields
 *   at the offsets 4 and 6. Note that the sector size is defined as
 *   NTFS_SECTOR_SIZE and not as the hardware sector size (this is concordant
 *   with what the Windows NTFS driver does).
 *
 * This function performs these 2 checks, and _fails_ if:
 * . the input size is invalid
 * . the fixup header is invalid
 * . the size does not match the number of sectors
 * . the magic identifier is wrong
 * . a fixup is invalid
 */
rf_int_t ntfs_fixup_record(rf_ntfs_volume_ref_t vol, rf_char_t* magic)
{
    rf_u32_t start, count, offset;
    rf_u16_t fixup;

    if (!IS_MAGIC(vol->sector_buffer, magic))
        return RF_ENOFS;
    start = u16_read(vol->sector_buffer, 4);
    count = u16_read(vol->sector_buffer, 6) - 1;
    if (vol->mft_record_size & (NTFS_SECTOR_SIZE - 1) || start & 1
            || start + count * 2 > vol->mft_record_size || vol->mft_record_size >> 9 != count) {
        if (vol->mft_record_size <= 0)
            return RF_ERROR;
        /*"NTFS: BUG: ntfs_fixup_record() got zero size! Please report this to "
        "linux-ntfs-dev@lists.sf.net\n"*/
    }
    fixup = u16_read(vol->sector_buffer, (rf_u16_t)start);
    start += 2;
    offset = NTFS_SECTOR_SIZE - 2;
    while (count--) {
        if (u16_read(vol->sector_buffer, (rf_u16_t)offset) != fixup)
            return RF_ERROR;
        u16_write(vol->sector_buffer, (rf_u16_t)offset, u16_read(vol->sector_buffer, (rf_u16_t)start));
        start += 2;
        offset += NTFS_SECTOR_SIZE;
    }
    return RF_ESUCCESS;
}

/* ҵmft_sectorƫ ش̻ƫ */
rf_int_t ntfs_read_attr(rf_ntfs_volume_ref_t vol, rf_u64_t sector, rf_u8_t need_read,
                        rf_u8_t attr, rf_wchar_t* attr_name,  rf_u16_t* Idx)
{
    rf_int_t ret;
    rf_u32_t type, len;
    rf_u16_t offset, name_len, name_offset, atrr_name_len;

    /* ǷҪݵ */
    if (need_read) {
        ret = rf_disk_read(vol->dev_handle, sector, vol->mft_sizeof_sector, vol->sector_buffer);
        if (ret)
            return ret;
        ret = ntfs_fixup_record(vol, "FILE");
        if (ret)
            return ret;
    }

    if (attr == AT_FILE_HEAD) {
        /*ȡļͷ */
        *Idx = 0;
        return RF_ESUCCESS;
    }

    if (attr_name)
        atrr_name_len = (rf_int_t)rf_wstrlen(attr_name);
    offset = *Idx;
    if (*Idx == 0)
        offset = u16_read(vol->sector_buffer, 0x14);
    do {
        type = u32_read(vol->sector_buffer, offset);
        /* ͱȽϼȽ */
        if (attr == type) {
            name_len = u8_read(vol->sector_buffer, offset + 9);
            name_offset = u8_read(vol->sector_buffer, offset + 10);
            /*Ϊ0 û*/
            if (!attr_name) {
                if (!name_len) {
                    *Idx = offset;
                    return RF_ESUCCESS;
                }
            } else {
                if (atrr_name_len == name_len) {
                    if (!rf_wstrncmp_buf(vol->sector_buffer, (rf_u16_t)(offset + name_offset), attr_name, name_len)) {
                        *Idx = offset;
                        return RF_ESUCCESS;
                    }
                }
            }
        }
        /* ûҵ ش */
        if (type > attr)
            return RF_ENOENT;
        len = u32_read(vol->sector_buffer, (rf_u16_t)(offset + 4));
        offset += (rf_u16_t)len;
    } while (type != 0xffffffff);
    return RF_ERROR;
}

/* read_clust_chain MFTʱ 
mft_sector  MFTʼַ
need_read   ǷҪ´Ӵ
attr        
attr_name    UNICODE 
read_offset ȡֵƫ

*/
rf_int_t read_clust_chain(rf_ntfs_file_ref_t file, rf_u64_t sector, rf_u8_t need_read,
                          rf_u8_t attr, rf_wchar_t* attr_name, rf_u16_t* read_offset)
{
    rf_u32_t i, data_run_len, data_attr_len, offset;
    rf_u16_t data_attr_offset, count_offset;
    rf_u8_t cb_len, cb_offset;
    rf_s32_t last_clust;

    /* ȡһδ */
    data_attr_offset = 0;
    i = ntfs_read_attr(file->vol, sector, need_read, attr, attr_name, &data_attr_offset);
    if (i)
        return i;

    /* õʱ */
    data_attr_len = u32_read(file->vol->sector_buffer, (rf_u16_t)(data_attr_offset + 0x04));
    data_run_len = data_attr_len - 0x40;

    count_offset = *read_offset;
    /* г0 ȡƫƴг 0 */
    if (!data_run_len || count_offset >= data_run_len) {
        chain_init(&file->clust, TRUE);
        return RF_ESUCCESS;
    }

    /* ȡȡƫ */
    offset = data_attr_offset + u16_read(file->vol->sector_buffer, (rf_u16_t)(data_attr_offset + 0x20)) + count_offset;
    i = u8_read(file->vol->sector_buffer, offset);
    if (!i) {
        /* MFTʱ ռ */
        *read_offset = USHRT_MAX;
        return RF_ENDFILE;
    }
    cb_len = GET_CB_LEN((rf_u8_t)i);
    cb_offset = GET_CB_OFFSET((rf_u8_t)i);

    /* жǷʱʼȡغ */
    last_clust = 0;
    if (count_offset)
        last_clust = chain_get_last_clust(&file->clust);

    if (cb_offset) {
        file->clust.chain[0].cluster = last_clust +
                                       get_s32(file->vol->sector_buffer, (rf_u16_t)(offset + 1 + cb_len), (rf_u8_t)cb_offset);
    } else {
        file->clust.chain[0].cluster = NULL_CLUSTER;
    }
    file->clust.chain[0].number = get_s32(file->vol->sector_buffer, (rf_u16_t)(offset + 1), cb_len);

    chain_init(&file->clust, FALSE);
    /* õһԿʼƫƵַ */
    file->clust.last_chain_index = 1;
    last_clust = file->clust.chain[0].cluster;

    /* дݶȡƫ */
    count_offset += (cb_len + cb_offset + 1);
    offset += (cb_len + cb_offset + 1);
    while (1) {
        /* 00  */
        i = u8_read(file->vol->sector_buffer, offset);
        if (!i) {
            /* MFTʱ ռ */
            *read_offset = USHRT_MAX;
            return RF_ESUCCESS;
        }

        cb_len = GET_CB_LEN((rf_u8_t)i);
        cb_offset = GET_CB_OFFSET((rf_u8_t)i);
        if (cb_offset) {
            file->clust.chain[ file->clust.last_chain_index ].cluster
                = last_clust + get_s32(file->vol->sector_buffer, (rf_u16_t)(offset + 1 + cb_len), (rf_u8_t)cb_offset);
            last_clust = file->clust.chain[ file->clust.last_chain_index ].cluster;
        } else {
            file->clust.chain[ file->clust.last_chain_index ].cluster = NULL_CLUSTER;
        }

        file->clust.chain[ file->clust.last_chain_index ].number =
            get_s32(file->vol->sector_buffer, (rf_u16_t)(offset + 1), cb_len);
        /*  */
        file->clust.last_chain_index++;

        count_offset += (cb_len + cb_offset + 1);
        offset += (cb_len + cb_offset + 1);
        /* ֹȡݳʱܳ */
        if (count_offset > data_run_len)
            break;

        if (file->clust.last_chain_index == RF_MAX_CHAIN_LENGTH)
            break;
    }
    *read_offset = count_offset;
    return RF_ESUCCESS;
}

/*
̻ ָ
attr                ΪҪбеֵ
attrib_name         ΪԵ
start_offset        ΪʼƫƵַ
end_offset          Ϊַ
*byte_offset        Ϊбƫ
*mft_attrib_sector  ΪҵMFT
*/

/* бMFT в */
static rf_int_t mft_check_attrib_list(rf_ntfs_file_ref_t pfile, rf_u8_t attr, rf_wchar_t* attrib_name,
                                      rf_u16_t start_offset, rf_u16_t end_offset, rf_u64_t* byte_offset, rf_u64_t* mft_attrib_sector)
{
    rf_u64_t start_mft;
    rf_u32_t type;
    rf_u16_t attr_len, offset, atrr_name_len, name_offset;
    rf_u8_t name_len;

    offset = start_offset;
    atrr_name_len = (rf_int_t)rf_wstrlen(attrib_name);
    while (1) {
        /* offset ƫ8ı߽ */
        type = u32_read(pfile->vol->sector_buffer, offset);
        attr_len = u16_read(pfile->vol->sector_buffer, (rf_u16_t)(offset + 4));
        if (type == attr) {
            /* ҵλ */
            name_len = u8_read(pfile->vol->sector_buffer, offset + 6);
            name_offset = u8_read(pfile->vol->sector_buffer, offset + 7);
            start_mft = u64_read(pfile->vol->sector_buffer, (rf_u16_t)(offset + 0x10));

            /*Ϊ0 û*/
            if (!attrib_name  && !name_len) {
_NEXT:
                *mft_attrib_sector = mft2sec(pfile->vol, start_mft);
                /* дƫֵ һԿʼλ */
                *byte_offset = (offset + attr_len);
                return RF_ESUCCESS;
            }

            if (atrr_name_len == name_len) {
                /* UNICODEַȽ */
                name_offset += offset;
                if (!rf_wstrncmp_buf(pfile->vol->sector_buffer, name_offset, attrib_name, name_len))
                    goto _NEXT;
            }
        }

        /* Ҳ ش */
        if (type > attr)
            break;

        if (offset > end_offset)
            break;
        offset += attr_len;
    }
    return RF_ENOENT;
}

/* Ѵݶ */
rf_int_t read_clust_chain_data(rf_ntfs_file_ref_t file, rf_void_t* buf, rf_u32_t* len, rf_int_t read)
{
    rf_int_t ret;
    rf_u32_t first_byte, read_byte;
    rf_u32_t read_sector, count_sector, seq_sector;
    rf_u16_t last_byte;

    /* һζȡϴʣβ */
    first_byte = 0;
    /* ¼ȡֽ */
    read_byte = 0;

    /* ȫ ûдؿɶʱ ش */
    if (file->clust.chain_offset >= file->clust.last_chain_index)
        return RF_ERROR;

    if (file->clust.byte_offset) {
        /* Ҫȡ */
        first_byte = min(file->vol->sector_size - file->clust.byte_offset, (rf_u16_t) * len);
        if (read) {
            if (file->clust.chain[ file->clust.chain_offset ].cluster != NULL_CLUSTER) {
                ret = rf_disk_read(file->vol->dev_handle, (((file->clust.chain[file->clust.chain_offset].cluster
                                   + file->clust.cluster_offset) << file->vol->sectors_per_cluster_bits)
                                   + file->clust.sector_offset), 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                memcpy(buf, file->vol->sector_buffer + file->clust.byte_offset, first_byte);
            } else {
                /* ϡļݶȡ 0 */
                memset(buf, 0, first_byte);
            }
        }

        /* дƫ ǷҪ ƫ */
        file->clust.byte_offset += (rf_u16_t)first_byte;
        if (file->clust.byte_offset == file->vol->sector_size) {
            file->clust.byte_offset = 0;
            /* ƫƵֱӴ */
            chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, 1);
        }

        /* Ҫֽ سɹ */
        read_byte = first_byte;
        if (read_byte == *len)
            return RF_ESUCCESS;

        /* дѾȡֽ */
        *len = read_byte;
        /* Ѿ Ҫ ش */
        if (file->clust.chain_offset >= file->clust.last_chain_index)
            return RF_ERROR;
    }

    /* õȡβֽ */
    count_sector = (*len - first_byte) >> file->vol->sector_size_bits;
    last_byte = (rf_u16_t)((*len - first_byte) % file->vol->sector_size);
    /* ȡ */
    while (count_sector) {
        if (file->clust.chain_offset >= file->clust.last_chain_index)
            return RF_ENDFILE;

        /* õ */
        seq_sector = ((file->clust.chain[ file->clust.chain_offset ].number - file->clust.cluster_offset)
                      << file->vol->sectors_per_cluster_bits) - file->clust.sector_offset;
        read_sector = min(seq_sector, count_sector);

        /* һֻܶȡ65535 */
        if (read) {
            if (file->clust.chain[ file->clust.chain_offset ].cluster != NULL_CLUSTER) {
                ret = rf_disk_read(file->vol->dev_handle, (((file->clust.chain[file->clust.chain_offset].cluster
                                   + file->clust.cluster_offset) << file->vol->sectors_per_cluster_bits) + file->clust.sector_offset),
                                   (rf_u16_t)read_sector, (rf_u8_t*)buf + read_byte);
                if (ret)
                    return ret;
            } else {
                /* ϡļݶȡ 0 */
                memset(buf, 0, (rf_u32_t)(read_sector << file->vol->sector_size_bits));
            }
        }

        /* ȡֽ */
        read_byte += (rf_u32_t)(read_sector << file->vol->sector_size_bits);
        /* д֮ȡֽ */
        *len = read_byte;
        /*  ʣҪȡ */
        if (count_sector >= read_sector)
            count_sector -= read_sector;

        /* ƫ */
        chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, read_sector);
    }

    /* ͷ Ȼݵ */
    if (last_byte) {
        if (file->clust.chain_offset >= file->clust.last_chain_index)
            return RF_ERROR;

        if (read) {
            if (file->clust.chain[ file->clust.chain_offset ].cluster != NULL_CLUSTER) {
                ret = rf_disk_read(file->vol->dev_handle, (((file->clust.chain[file->clust.chain_offset].cluster
                                   + file->clust.cluster_offset) << file->vol->sectors_per_cluster_bits)
                                   + file->clust.sector_offset), 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                memcpy((rf_u8_t*)buf + read_byte, file->vol->sector_buffer, last_byte);
            } else {
                memset((rf_u8_t*)buf + read_byte, 0, last_byte);
            }
        }

        read_byte += last_byte;
        *len = read_byte;

        /* дƫֽ */
        file->clust.byte_offset += last_byte;
    }
    return RF_ESUCCESS;
}

/* б̻ */
static rf_int_t read_list_to_disk_buff(rf_ntfs_file_ref_t file)
{
    rf_int_t i;
    rf_u32_t read_len;

    /* һֽݵ̻ */
    read_len = file->vol->sector_size;

    /* ռ ռ */
    if (file->clust.last_chain_index == 0 || file->clust.chain_offset >= file->clust.last_chain_index) {
        i = read_clust_chain(file, file->entry.sector, TRUE, AT_ATTRIBUTE_LIST, NULL, &file->list_body_offset);
        if (i)
            return i;
    }
    return read_clust_chain_data(file, file->vol->sector_buffer, &read_len, TRUE);
}

static rf_int_t check_attrib_list(rf_ntfs_file_t* file, rf_u8_t attr, rf_wchar_t* attrib_name,
                                  rf_u16_t start_offset, rf_u16_t end_offset, rf_u64_t* byte_offset, rf_u64_t* mft_attrib_sector)
{
    rf_int_t ret;
    rf_u64_t start_mft;
    rf_u32_t type, offset_sector;
    rf_u16_t attr_len, offset, atrr_name_len, name_offset, t, a;
    rf_u8_t name_len;
    rf_u8_t read_next_bit;

    offset = start_offset;
    atrr_name_len = (rf_int_t)rf_wstrlen(attrib_name);
    offset_sector = 0;
    while (1) {
        /* offset ƫ8ı߽ */
        type = u32_read(file->vol->sector_buffer, offset);
        /*  */
        if (!type)
            break;

        attr_len = u16_read(file->vol->sector_buffer, (rf_u16_t)(offset + 4));
        if (type == attr) {
            /* ҵλ */
            name_len = u8_read(file->vol->sector_buffer, offset + 6);
            name_offset = u8_read(file->vol->sector_buffer, offset + 7);

            /* Ƿһ־λ */
            read_next_bit = FALSE;

            /* ǷԽ߽ */
            t = offset + 0x10;
            /* бʱ ÿζȡһ ڴ˼ƫ */
            if (t >= file->vol->sector_size) {
                read_next_bit = TRUE;
                ret = read_list_to_disk_buff(file);
                if (ret)
                    return ret;
                /* õMFTڵڶƫ */
                t %= file->vol->sector_size;
                start_mft = u64_read(file->vol->sector_buffer, t);
            }
            start_mft = u64_read(file->vol->sector_buffer, t);

            /*Ϊ0 û*/
            if (!attrib_name  && !name_len) {
_NEXT:
                /* ԵMFT vol->mft_sizeof_sector2 Ϊ512ֽ */
                *mft_attrib_sector = mft2sec(file->vol, start_mft);

                /* дƫֵ һԿʼλ */
                *byte_offset += ((offset_sector << file->vol->sector_size_bits) + (offset + attr_len));
                return RF_ESUCCESS;
            }

            if (attrib_name) {
                if (atrr_name_len == name_len) {
                    /* Ƿȫ MFT */
                    if ((offset + name_offset + name_len) <= file->vol->sector_size) {
                        /* UNICODEַȽ ǷԽ߽ */
                        name_offset += offset;
_COMPARE:
                        if (!rf_wstrncmp_buf(file->vol->sector_buffer, name_offset, attrib_name, name_len))
                            goto _NEXT;
                    } else {
                        /* ֿܿԽ ȫڵڶ*/
                        if ((offset + name_offset) >= file->vol->sector_size) {
                            if ((offset + name_offset + name_len) <= (file->vol->sector_size << 1)) {
                                /* ļȫڵڶ */
                                if (!read_next_bit) {
                                    ret = read_list_to_disk_buff(file);
                                    if (ret)
                                        return ret;
                                    name_offset += offset;
                                    name_offset %= file->vol->sector_size;
                                }
                                /* ԵMFT Ѿȡ */
                                goto _COMPARE;
                            } else {
                                /* ļ ڵĵڶ  */
                                if (!read_next_bit) {
                                    /* ûжȡ ȡ */
                                    ret = read_list_to_disk_buff(file);
                                    if (ret)
                                        return ret;
                                }

                                /* ڶ ʣ */
                                t = ((file->vol->sector_size << 1) - (offset + name_offset)) >> 1;
                                /* ڶƫƵַ */
                                a = (offset + name_offset) % file->vol->sector_size;
                                if (!rf_wstrncmp_buf(file->vol->sector_buffer, a, attrib_name, t)) {
                                    ret = read_list_to_disk_buff(file);
                                    if (ret)
                                        return ret;
                                    a = ((offset + name_offset + name_len) - file->vol->sector_size) >> 1;
                                    if (!rf_wstrncmp_buf(file->vol->sector_buffer, 0, attrib_name + t, a))
                                        goto _NEXT;
                                }
                            }
                        } else {
                            /* ļڵһ ڶ */
                            /* Ƚϵһʣ */
                            t = (file->vol->sector_size - (offset + name_offset)) >> 1;
                            if (!rf_wstrncmp_buf(file->vol->sector_buffer, (rf_u16_t)(offset + name_offset), attrib_name, t)) {
                                ret = read_list_to_disk_buff(file);
                                if (ret)
                                    return ret;

                                /* Ƚϵڶʣµ */
                                a = ((offset + name_offset + name_len) - file->vol->sector_size) >> 1;
                                if (!rf_wstrncmp_buf(file->vol->sector_buffer, 0, attrib_name + t, a))
                                    goto _NEXT;
                            }
                        }
                    }
                }
            }
        }
        /* Ҳ ش */
        if (type > attr)
            return RF_ENOENT;

        offset += attr_len;
        /*  ƫ*/
        if (offset >= file->vol->sector_size) {
            offset_sector++;
            /* if( offset % file->vol->sector_size == 0 ) */
            {   /* ļԳȸպ߽ һ */
                ret = read_list_to_disk_buff(file);
                if (ret)
                    return ret;
            }
            offset %= file->vol->sector_size;
        }
    }
    return RF_ENOENT;
}

/*
б õƫڵMFT
attr                ΪҪбеֵ
attrib_name         ΪԵ
attr_index          ͬ
mft_attrib_sector   زMFTַ
*/
rf_int_t ntfs_check_attrib_list(rf_ntfs_file_ref_t file, rf_u8_t attr, rf_wchar_t* attrib_name,
                                rf_u64_t* search_offset, rf_u64_t* mft_attrib_sector)
{
    rf_int_t ret;
    rf_u32_t offset_byte;
    rf_u16_t idx_offset;
    rf_u32_t count_sector, temp_sector;

    if (!file->list_run_bit) {
        /* б Ҫнһķ!!! */
        idx_offset = 0;
        ret = ntfs_read_attr(file->vol, file->entry.sector, TRUE, AT_ATTRIBUTE_LIST, NULL, &idx_offset);
        if (ret == RF_ESUCCESS) {
            /* ʼļ ȡ׼MFT */
            rf_u32_t list_offset, list_max_offset;

            list_offset = idx_offset + u16_read(file->vol->sector_buffer, (rf_u16_t)(idx_offset + 0x14));
            list_max_offset = list_offset + u32_read(file->vol->sector_buffer, (rf_u16_t)(idx_offset + 0x10));
            /* ƫ ش */
            if (*search_offset >  list_max_offset)
                return RF_ERROR;

            /* һδӿʼַʼ */
            if (*search_offset)
                list_offset = (rf_u32_t)(*search_offset);

            /* Խв */
            ret = mft_check_attrib_list(file, attr, attrib_name, (rf_u16_t)list_offset,
                                        (rf_u16_t)list_max_offset, search_offset, mft_attrib_sector);
            return ret;
        }
    } else {
        /* б ÿδͷʼռ!!! */
        chain_init(&file->clust, TRUE);
        /* ͷʼռ */
        file->list_body_offset = 0;

        /* ʼֽڼƫ */
        count_sector = (rf_u32_t)(*search_offset >> file->vol->sector_size_bits);
        while (1) {
            if (file->clust.last_chain_index == 0) {
_CONTI:
                ret = read_clust_chain(file, file->entry.sector, TRUE, AT_ATTRIBUTE_LIST, NULL, &file->list_body_offset);
                if (ret)
                    return ret;
            }

            if (count_sector == 0)
                break;

            /* 㵱ǰؿ */
            temp_sector = chain_count_clust(&file->clust) << file->vol->sectors_per_cluster_bits;
            if (count_sector >= temp_sector) {
                count_sector -= temp_sector;
                goto _CONTI;
            } else {
                chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, count_sector);
                break;
            }
        }

        /* һ̻ */
        ret = read_list_to_disk_buff(file);
        if (ret)
            return ret;

        offset_byte = (rf_u16_t)((*search_offset) % file->vol->sector_size);
        /* д *search_offset ƫֽڵı */
        *search_offset = (count_sector << file->vol->sector_size_bits);
        /* ݵ Ȼз */
        ret = check_attrib_list(file, attr, attrib_name, (rf_u16_t)offset_byte,
                                (rf_u16_t)file->vol->sector_size, search_offset, mft_attrib_sector);
    }
    return ret;
}

rf_int_t check_index_bitmap(rf_ntfs_file_ref_t file, rf_bool_t* is_use_bit)
{
    rf_int_t ret;
    rf_wchar_t name[ 5 ];
    rf_u16_t offset, idx_offset;
    rf_u32_t i, len;

    idx_offset = 0;
    rf_a2w_strcpy(name, "$I30");
    ret = ntfs_read_attr(file->vol, file->u.di.index_bitmap_sector, TRUE, AT_BITMAP, name, &idx_offset);
    if (ret)
        return ret;

    /* bitmapʱ!! */
    if (u8_read(file->vol->sector_buffer, idx_offset + 8) == 1) {
        /* ֧bitmapʱ ֱӷ ǰؿ!! */
        *is_use_bit = TRUE;
        return RF_ESUCCESS;
    }

    /* õƫ */
    offset = idx_offset + u16_read(file->vol->sector_buffer, (rf_u16_t)(idx_offset + 0x14));
    /* õ */
    len = u16_read(file->vol->sector_buffer, (rf_u16_t)(idx_offset + 0x10));

    /* ൱ڳ8 õֽƫ */
    i = file->u.di.enum_vcn_index >> 3;
    /* ļ */
    if (i >= len)
        return RF_ENDFILE;

    /* ʱñlen */
    len = u8_read(file->vol->sector_buffer, offset + i);
    /* õλƫ */
    i = file->u.di.enum_vcn_index % 8;

    *is_use_bit = GET_BIT(len, i);
    /* ¿ʼӦVCNǷ!!! */
    return RF_ESUCCESS;
}

rf_int_t read_fixup_indx_sector(rf_ntfs_file_ref_t file)
{
    rf_int_t ret;
    rf_u32_t i, len;
    rf_u16_t data[ INDX_REPAIR_NUM ], fixup, count, count_bits, offset;

    /* ʼص */
    ret = rf_disk_read(file->vol->dev_handle,
                       file->u.di.enum_record_last_cluster << file->vol->sectors_per_cluster_bits,
                       1, file->vol->sector_buffer);
    if (ret)
        return ret;

    /* ڴ˵õֵ */
    fixup = u16_read(file->vol->sector_buffer, 0x28);
    /* Ҫȡֵ */
    count = file->vol->sector_size >> NTFS_SECTOR_BITS;
    count_bits = rf_ffs(count);
    for (i = 0; i < count; i++) {
        data[i] = u16_read(file->vol->sector_buffer,
                           (rf_u16_t)(0x2a + ((file->u.di.enum_record_sector_offset << 1) << count_bits) + (i << 1)));
    }

    /* жϵǰǷʼ */
    i = file->clust.chain[ file->clust.chain_offset ].cluster + file->clust.cluster_offset + file->clust.sector_offset;
    if (i != file->u.di.enum_record_last_cluster) {
        /* ÿζȡһ */
        len = file->vol->sector_size;
        /* Ѵݶ */
        ret = read_clust_chain_data(file, file->vol->sector_buffer, &len, TRUE);
        /*  F_ERR ΪƫƵβ ¶´ص */
        if (ret)
            return ret;
    } else {
        /* ֱǰƫ */
        chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, 1);
    }

    /* ٴνжԱȼ */
    offset = NTFS_SECTOR_SIZE - 2;
    i = 0;
    while (count--) {
        if (u16_read(file->vol->sector_buffer, (rf_u16_t)offset) != fixup)
            return RF_ERROR;
        u16_write(file->vol->sector_buffer, offset, (rf_u16_t)data[i]);

        i++;
        offset += NTFS_SECTOR_SIZE;
    }
    return RF_ESUCCESS;
}

rf_int_t read_dir_to_disk_buf(rf_ntfs_file_ref_t file)
{
    rf_int_t ret;
    rf_wchar_t name[ 5 ];

    /* ôƫ */
    if (file->clust.last_chain_index == 0 || file->clust.chain_offset >= file->clust.last_chain_index) {
        /*  */
        rf_a2w_strcpy(name, "$I30");
        ret = read_clust_chain(file, file->u.di.index_alloc_sector, TRUE, AT_INDEX_ALLOCATION,
                               name, &file->u.di.index_alloc_body_offset);
        if (ret)
            return ret;
    }
    /* ¼¼ı */
    file->u.di.enum_record_sector_offset++;
    return read_fixup_indx_sector(file);
}

/* ƫoffset_clustغ Ŀ¼ݵ */
rf_int_t dir_read_offset_to_disk_buf(rf_ntfs_file_ref_t file, rf_u32_t offset_clust)
{
    rf_int_t ret;
    rf_wchar_t name[ 5 ];

    /* һռ */
    if (file->clust.last_chain_index == 0)
        goto _NEXT;

    while (1) {
        if (file->clust.chain_offset >= file->clust.last_chain_index) {
_NEXT:
            rf_a2w_strcpy(name, "$I30");
            if (file->u.di.index_alloc_body_offset != USHRT_MAX) {
                ret = read_clust_chain(file, file->u.di.index_alloc_sector, TRUE,
                                       AT_INDEX_ALLOCATION, name, &file->u.di.index_alloc_body_offset);
                if (ret)
                    return ret;
            } else {
                /* ######### */
                ret = ntfs_check_attrib_list(file, AT_INDEX_ALLOCATION, name,
                                             &file->list_search_offset, &file->u.di.index_alloc_sector);
                if (ret)
                    return ret;

                file->u.di.index_alloc_body_offset = 0;
                /* ٴλȡõ */
                ret = read_clust_chain(file, file->u.di.index_alloc_sector, TRUE,
                                       AT_INDEX_ALLOCATION, name, &file->u.di.index_alloc_body_offset);
                if (ret)
                    return ret;
            }
        }

        if (offset_clust == 0)
            break;
        offset_clust--;
        chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, file->vol->sectors_per_cluster);
    }
    /* õĿ¼ṹĿʼ */
    file->u.di.enum_record_last_cluster = file->clust.chain[ file->clust.chain_offset ].cluster + file->clust.cluster_offset;
    return read_fixup_indx_sector(file);
}
