/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "ntfs_volume.h"
#include "ntfs_file.h"
#include "ntfs.h"
#include "../src/rf_fs.h"

#define RF_COMM_MAX_VOLUME             RF_OPT_NTFS_MAX_VOLUME
#define RF_COMM_MAX_FILE               RF_OPT_NTFS_MAX_FILE
#define RF_COMM_INIT_FS_TYPE           RF_TYPE_NTFS
#define rf_volume_comm_t               rf_ntfs_volume_t
#define rf_volume_comm_ref_t           rf_ntfs_volume_ref_t
#define rf_file_comm_t                 rf_ntfs_file_t
#define rf_file_comm_ref_t             rf_ntfs_file_ref_t
#define comm_get_volume_handle_count   ntfs_get_volume_handle_count
#define comm_get_file_handle_count     ntfs_get_file_handle_count
#define comm_check_file_access         ntfs_check_file_access
#define init_data                      ntfs_init
#define vol_alloc                      ntfs_vol_alloc
#define vol_free                       ntfs_vol_free
#include "../common/rf_comdata.h"

rf_int_t ntfs_mount(struct rf_vol_ctx* vctx, rf_mount_param_t* param)
{
    rf_ntfs_volume_ref_t vol;
    rf_int_t ret;

    vol = ntfs_vol_alloc();
    if (vol == NULL)
        return RF_EMFILE;

    ret = ntfs_volume_open(vol, param);
    if (ret == RF_ESUCCESS) {
        vctx->vol = vol;
        vol->open_flag = 1;
        return RF_ESUCCESS;
    }

    if ((rf_vol_comm_ref_t)vol->dev_handle != NULL)
        rf_disk_close((rf_vol_comm_ref_t)vol->dev_handle);

#if RF_OPT_USE_MALLOC == 1
    if (vol->sector_buffer != NULL)
        rf_free(vol->sector_buffer);
#endif
    ntfs_vol_free(vol);
    return ret;
}

rf_int_t ntfs_unmount(struct rf_vol_ctx* vctx)
{
    rf_int_t ret;

    ret = ntfs_volume_close(vctx->vol);
    if (ret == RF_ESUCCESS)
        ntfs_vol_free(vctx->vol);
    return ret;
}

rf_int_t ntfs_ioctl_vol(struct rf_vol_ctx* vctx, rf_int_t cmd, rf_void_t* args)
{
    switch (cmd) {
    case RF_VOL_IOCTL_STATFS:
        return ntfs_volume_stat(vctx->vol, (rf_statfs_ref_t)args);
    }
    return  RF_EPERM;
}

#if RF_OPT_USE_RAWOPEN
rf_int_t ntfs_raw_open(struct rf_file_ctx* fd, rf_fentry_t* fentry, rf_int_t flags)
{
    rf_int_t ret;
    rf_ntfs_file_ref_t file;

    file = file_alloc((rf_ntfs_volume_ref_t)fd->vctx->vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = ntfs_file_raw_open(file, fentry, flags);
    if (ret != RF_ESUCCESS) {
        if (ret == RF_ENDFILE)
            ret = RF_ENOENT;
        file_free(file);
        file = NULL;
    }
    fd->file = file;
    return ret;
}
#endif

rf_int_t ntfs_open(struct rf_file_ctx* fd, const rf_char_t* file_name, rf_int_t flags)
{
    rf_ntfs_file_ref_t file;
    rf_int_t ret;

    file = file_alloc((rf_volume_comm_ref_t)fd->vctx->vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = ntfs_file_open(file, file_name, flags);
    if (ret != RF_ESUCCESS) {
        if (ret == RF_ENDFILE)
            ret = RF_ENOENT;
        file_free(file);
        file = NULL;
    }
    fd->file = file;
    return ret;
}

rf_int_t ntfs_close(struct rf_file_ctx* fd)
{
    rf_int_t ret;

    ret = ntfs_file_close((rf_file_comm_ref_t)fd->file);
    if (!ret)
        file_free((rf_file_comm_ref_t)fd->file);
    return ret;
}

rf_int_t ntfs_ioctl(struct rf_file_ctx* fd, rf_int_t cmd, rf_void_t* args)
{
    switch (cmd) {
    case RF_FD_IOCTL_GET_FSTAT:
        return ntfs_file_get_stat(fd->file, (rf_stat_ref_t)args);
#if RF_OPT_USE_GET_FTIME
    case RF_FD_IOCTL_GET_FTIME:
        return ntfs_file_get_ftime(fd->file, (rf_file_ftime_ref_t)args);
#endif
    }
    return RF_EPERM;
}

#if RF_OPT_USE_READDIR
rf_int_t ntfs_readdir(struct rf_file_ctx* fd, const rf_char_t* pattern, rf_dirent_t* dirp)
{
    return ntfs_file_read_dir(fd->file, pattern, dirp);
}
#endif

#if RF_OPT_USE_FSEEK
rf_int_t ntfs_lseek(struct rf_file_ctx* fd, rf_fsize_t offset, rf_int_t origin, rf_fsize_t* now_offset)
{
    rf_int_t ret;

    ret = ntfs_file_lseek(fd->file, (rf_u64_t)offset, origin);
    if (ret)
        return ret;
    *now_offset = ((rf_ntfs_file_ref_t)fd->file)->current_offset;
    return 0;
}
#endif

rf_int_t ntfs_read(struct rf_file_ctx* fd, rf_void_t* buf, rf_size_t len, rf_size_t* plen)
{
    return ntfs_file_read(fd->file, buf, (rf_u32_t)len, (rf_u32_t*)plen);
}
