/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_dos.h"
#include <string.h>

#define DOS_MBR                     0
#define DOS_PBR                     1
#define is_extended(t)              (t == 0x5 || t == 0xf || t == 0x85)

struct rf_part_ctx {
    rf_handle_t dev;
    rf_u8_t* buffer;
    rf_part_param_t* param;
};
typedef struct rf_part_ctx rf_part_ctx_t, *rf_part_ctx_ref_t;

#if RF_OPT_USE_MBR
static rf_int_t test_block_type(rf_u8_t* buffer)
{
    rf_int_t slot;
    rf_u8_t boot_ind;

    if ((buffer[DOS_PART_MAGIC_OFFSET + 0] != 0x55) ||
            (buffer[DOS_PART_MAGIC_OFFSET + 1] != 0xaa)) {
        return RF_ENOFS;
    }

    for (slot = 0; slot < 3; slot++) {
        boot_ind = buffer[DOS_PART_TBL_OFFSET + (slot * 16 + DOS_PART_BOOT_IND_OFFSET) ];
        if (boot_ind != 0 && boot_ind != 0x80) {
            if (!slot) {
                return DOS_PBR;
            } else {
                return RF_ENOFS;
            }
        }
    }

    if (*((rf_u32_t*)&buffer[DOS_PART_TBL_OFFSET + (0 * 16 + DOS_PART_START_OFFSET)]) == 0 &&
            *((rf_u32_t*)&buffer[DOS_PART_TBL_OFFSET + (1 * 16 + DOS_PART_START_OFFSET)]) == 0 &&
            *((rf_u32_t*)&buffer[DOS_PART_TBL_OFFSET + (2 * 16 + DOS_PART_START_OFFSET)]) == 0 &&
            *((rf_u32_t*)&buffer[DOS_PART_TBL_OFFSET + (3 * 16 + DOS_PART_START_OFFSET)]) == 0)
        return DOS_PBR;
    return DOS_MBR;
}

static rf_int_t get_extended_parts(rf_part_ctx_t* ctx, rf_lloff_t start, rf_lloff_t relative)
{
    rf_int_t ret;
    rf_int_t i;
    rf_u8_t boot_ind, sys_ind;
    rf_lloff_t lba_start;

    ret = rf_disk_read(ctx->dev, start, 1, ctx->buffer);
    if (ret)
        return ret;

    if (ctx->buffer[DOS_PART_MAGIC_OFFSET] != 0x55 ||
            ctx->buffer[DOS_PART_MAGIC_OFFSET + 1] != 0xaa) {
        return RF_ENOFS;
    }

    for (i = 0; i < 4; i++) {
        /*
        * fdisk does not show the extended partitions that
        * are not in the MBR
        */
        boot_ind = ctx->buffer[(DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_BOOT_IND_OFFSET))];
        sys_ind = ctx->buffer[(DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_SYS_IND_OFFSET))];
        /*printf( "type=0x%x %d\n", sys_ind, (DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_SYS_IND_OFFSET)));*/
        if ((sys_ind != 0) && (is_extended(sys_ind) == 0)) {

            if (ctx->param->count < RF_MAX_DISK_PART) {
                ctx->param->part[ctx->param->count].start =
                    (rf_lloff_t)(start + u32_read(ctx->buffer, DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_START_OFFSET)));
                ctx->param->part[ctx->param->count].size =
                    (rf_lloff_t)(start + u32_read(ctx->buffer, DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_OFFSET_OFFSET)));
            }
            ctx->param->count++;
        }
    }

    for (i = 0; i < 4; i++) {
        sys_ind = ctx->buffer[DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_SYS_IND_OFFSET)];
        if (is_extended(sys_ind)) {
            lba_start = u32_read(ctx->buffer, DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_START_OFFSET)) + relative;
            ret = get_extended_parts(ctx, lba_start, start == 0 ? lba_start : relative);
            if (ret)
                return ret;
        }
    }
    return 0;
}

rf_int_t dos_get_parts(rf_handle_t dev, rf_u8_t* buffer, rf_part_param_t* param)
{
    rf_int_t ret;
    rf_part_ctx_t ctx;

    ret = rf_disk_read(dev, 0, 1, buffer);
    if (ret)
        return ret;

    ret = test_block_type(buffer);
    if (ret == DOS_MBR) {
        /* ʼmbr  */
        ctx.dev = dev;
        ctx.buffer = buffer;
        ctx.param = param;
        return get_extended_parts(&ctx, 0, 0);
    } else {
        param->count = 1;
        param->part->start = 0;
        param->part->size = 0;
        return 0;
    }
    return ret;
}
#endif
