/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_file.h"
#include "rf_fd.h"

#if RF_OPT_USE_RF_PART
#include "../part/rf_part.h"
rf_int_t rf_part(rf_char_t* name, rf_part_param_t* param)
{
    rf_int_t ret;

    ret = rf_imp_part(name, param);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif


#if RF_OPT_USE_MKFS
rf_int_t rf_mkfs(const rf_char_t* fs_type, const rf_char_t* device_name, const rf_void_t* fc)
{
    rf_int_t ret;

    ret = rf_imp_mkfs(fs_type, device_name, fc);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

rf_int_t rf_mount(const rf_char_t* device_name, const rf_char_t* path,
                  const rf_char_t* fs_type, rf_int_t flags, const rf_mount_ctx_t* ctx)
{
    rf_int_t ret;
    rf_mount_param_t param;

    param.device_name = device_name;
    param.flags = flags;
    param.ctx = (rf_mount_ctx_t*)ctx;
    ret = rf_imp_mount(path, fs_type, &param);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}

rf_int_t rf_unmount(const rf_char_t* path)
{
    rf_int_t ret;

    ret = rf_imp_unmount(path);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}

#if RF_OPT_USE_STATFS
rf_int_t rf_statfs(const rf_char_t* path, struct rf_statfs* buf)
{
    rf_int_t ret;

    ret = rf_imp_statfs(path, buf);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_FILE_SECBUF
rf_void_t* rf_get_volume_sector_buf(const rf_char_t* path)
{
    rf_vol_ctx_t* fs;
    rf_vol_comm_ref_t vol;

    fs = find_fs_ctx(path);
    if (fs == NULL) {
        rf_set_errno(RF_EINVAL);
        return NULL;
    }
    vol = (rf_vol_comm_ref_t)fs->vol;
    return vol->sector_buffer;
}

rf_int_t rf_get_volume_sector_size(const rf_char_t* path)
{
    rf_statfs_t st;

    if (rf_statfs(path, &st) < 0)
        return -1;
    return st.f_ssize;
}
#endif

#if RF_OPT_USE_CHDIR
rf_int_t rf_chdir(const rf_char_t* path)
{
    rf_int_t ret;

    ret = rf_imp_chdir(path);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_GETCWD
rf_int_t rf_getcwd(rf_char_t* buf, rf_size_t size)
{
    rf_int_t ret;

    ret = rf_imp_getcwd(buf, size);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_RAWOPEN
rf_int_t rf_raw_open(const rf_char_t* path, rf_fentry_t* fentry, rf_int_t flags)
{
    rf_int_t fd, ret;
    rf_file_ctx_t* fc;

    fd = fd_new();
    if (fd < 0) {
        rf_set_errno(RF_EMFILE);
        return -1;
    }

    fc = fd_get_check(fd);
    ret = rf_file_raw_open(fc, path, fentry, flags);
    if (ret != 0) {
        fd_free(fc);
        rf_set_errno(ret);
        return -1;
    }
    return fd;
}
#endif

rf_int_t rf_open(const rf_char_t* pathname, rf_int_t flags)
{
    rf_int_t fd, ret;
    rf_file_ctx_t* fc;

    fd = fd_new();
    if (fd < 0) {
        rf_set_errno(RF_EMFILE);
        return -1;
    }

    fc = fd_get_check(fd);
    ret = rf_file_open(fc, pathname, flags);
    if (ret != 0) {
        fd_free(fc);
        rf_set_errno(ret);
        return -1;
    }
    return fd;
}

rf_int_t rf_close(rf_int_t fd)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_close(fc);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    fd_free(fc);
    return 0;
}

#if RF_OPT_USE_READDIR
rf_int_t rf_readdir_ex(rf_int_t fd, const rf_char_t* pattern, struct rf_dirent_i* ptr)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_readdir(fc, pattern, ptr);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

rf_int_t rf_ioctl(rf_int_t fd, rf_u32_t cmd, rf_void_t* buf)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(fc, cmd, buf);
    if (ret != RF_ESUCCESS) {
        rf_set_errno(ret);
        ret = -1;
    }
    return 0;
}

#if RF_OPT_USE_GET_FSTAT
rf_int_t rf_fstat(rf_int_t fd, rf_stat_t* buf)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    if (!buf)
        return RF_EINVAL;
    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(fc, RF_FD_IOCTL_GET_FSTAT, buf);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_FSEEK
rf_off_t rf_lseek(rf_int_t fd, rf_off_t offset, rf_int_t whence)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;
    rf_off_t now_offset;

    if (offset < 0)
        return RF_EINVAL;
    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_lseek(fc, offset, whence, &now_offset);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return now_offset;
}
#endif

rf_ssize_t rf_read(rf_int_t fd, rf_void_t* buf, rf_size_t count)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;
    rf_size_t rlen;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }
    ret = rf_file_read(fc, buf, count, &rlen);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return rlen;
}

#if RF_OPT_FS_READONLY == 0
rf_ssize_t rf_write(rf_int_t fd, const rf_void_t* buf, rf_size_t count)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;
    rf_size_t wlen;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_write(fc, buf, count, &wlen);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return wlen;
}

rf_int_t rf_fsync(rf_int_t fd)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(fc, RF_FD_IOCTL_FSYNC, NULL);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_RENAME
rf_int_t rf_rename(const rf_char_t* oldname, const rf_char_t* newname)
{
    rf_int_t ret;

    ret = rf_file_rename(oldname, newname);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_FDELETE
rf_int_t rf_fdelete(rf_int_t fd)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(fc, RF_FD_IOCTL_FDELETE, NULL);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_UNLINK
rf_int_t rf_unlink(const rf_char_t* pathname)
{
    rf_int_t ret;

    ret = rf_file_unlink(pathname);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_MKDIR
rf_int_t rf_mkdir(const rf_char_t* path)
{
    rf_int_t fd;
    rf_file_ctx_t* fc;
    rf_int_t ret;

    fd = fd_new();
    if (fd == -1) {
        rf_set_errno(RF_EMFILE);
        return -1;
    }

    fc = fd_get_check(fd);
    ret = rf_file_open(fc, path, RF_O_DIRECTORY | RF_O_CREAT);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }

    ret = rf_file_close(fc);
    fd_free(fc);
    return ret;
}
#endif

#if RF_OPT_USE_FTRUNCATE
rf_int_t rf_ftruncate(rf_int_t fd, rf_off_t length)
{
    rf_int_t ret;
    rf_file_ctx_t* fc;

    fc = fd_get_check(fd);
    if (fc == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(fc, RF_FD_IOCTL_FTRUNCATE, &length);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif
