/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_file.h"
#include "rf_fd.h"

#if RF_OPT_USE_GET_FNAME
rf_int_t rf_get_name(rf_int_t fd, rf_char_t* buf, rf_u32_t len)
{
    rf_int_t ret;
    rf_file_ctx_t* ctx;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(ctx, RF_MAKE_FD_IOCTL(RF_FD_IOCTL_GET_SHORT_NAME, len), buf);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_GET_SET_LONG_NAME
rf_int_t rf_get_long_name(rf_int_t fd, rf_wchar_t* buf, rf_int_t len)
{
    rf_int_t ret;
    rf_file_ctx_t* ctx;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(ctx, RF_MAKE_FD_IOCTL(RF_FD_IOCTL_GET_NAME, len), buf);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_GET_FTIME
rf_int_t rf_get_time(rf_int_t fd, rf_file_ftime_ref_t ft)
{
    rf_int_t ret;
    rf_file_ctx_t* ctx;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(ctx, RF_FD_IOCTL_GET_FTIME, ft);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_GET_LTIME
rf_int_t rf_get_ltime(rf_int_t fd, rf_tm_ref_t create, rf_tm_ref_t last_access, rf_tm_ref_t last_write)
{
    rf_int_t ret;
    rf_file_ctx_t* ctx;
    rf_file_ltime_t lt;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(ctx, RF_FD_IOCTL_GET_LTIME, &lt);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    if (create)
        *create = lt.create;
    if (last_access)
        *last_access = lt.last_access;
    if (last_write)
        *last_write = lt.last_write;
    return 0;
}
#endif

#if RF_OPT_USE_SET_FTIME
rf_int_t rf_set_time(rf_int_t fd, rf_ftime_ref_t create, rf_ftime_ref_t last_access, rf_ftime_ref_t last_write)
{
    rf_file_ctx_t* ctx;
    rf_file_ftime_t ft;
    rf_int_t ret;

    if (create == NULL || last_access == NULL || last_write == NULL) {
        ret = rf_get_time(fd, &ft);
        if (ret != RF_ESUCCESS)
            return ret;
    }
    if (create)
        ft.create = *create;
    if (last_access)
        ft.last_access = *last_access;
    if (last_write)
        ft.last_write = *last_write;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }
    ret = rf_file_ioctl(ctx, RF_FD_IOCTL_SET_FTIME, &ft);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_SET_LTIME
rf_int_t rf_set_ltime(rf_int_t fd, rf_tm_ref_t create, rf_tm_ref_t last_access, rf_tm_ref_t last_write)
{
    rf_file_ctx_t* ctx;
    rf_file_ltime_t lt;
    rf_int_t ret;

    if (create == NULL || last_access == NULL || last_write == NULL) {
        ret = rf_get_ltime(fd, &lt.create, &lt.last_access, &lt.last_write);
        if (ret != RF_ESUCCESS)
            return ret;
    }
    if (create)
        lt.create = *create;
    if (last_access)
        lt.last_access = *last_access;
    if (last_write)
        lt.last_write = *last_write;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }
    ret = rf_file_ioctl(ctx, RF_FD_IOCTL_SET_LTIME, &lt);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_SET_ATTRIB
rf_int_t rf_set_attrib(rf_int_t fd, rf_u32_t attrib)
{
    rf_int_t ret;
    rf_file_ctx_t* ctx;

    ctx = fd_get_check(fd);
    if (ctx == NULL) {
        rf_set_errno(RF_EBADF);
        return -1;
    }

    ret = rf_file_ioctl(ctx, RF_FD_IOCTL_SET_ATTRIB, &attrib);
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif

#if RF_OPT_USE_FILE_CTX
rf_int_t rf_set_file_ctx(rf_int_t fd, rf_void_ref_t ctx)
{
    return -1;
}

rf_int_t rf_get_file_ctx(rf_int_t fd, rf_void_ref_t* ctx)
{
    return -1;
}
#endif
