/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../codepage/rf_codepage.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_file.h"

struct rf_find_ctx {
    rf_int_t            fd;
    const rf_char_t*    pattern;
};
typedef struct rf_find_ctx rf_find_ctx_t, *rf_find_ctx_ref_t;

rf_handle_t rf_findfirst(const rf_char_t* path, struct rf_dirent* ptr)
{
    rf_char_t dir[RF_MAX_PATH];
    rf_find_ctx_t* fctx;
    const rf_char_t* p;
    rf_int_t fd, ret;

    p = path_get_filespec(path);
    if (!rf_str_contain_wild_cards(p)) {
        /*  *   */
        rf_set_errno(RF_EINVAL);
        return NULL;
    }
    rf_strcpy_s(dir, sizeof(dir), path);
    if (!path_remove_filespec(dir)) {
        /* ڵǰĿ¼ öļ */
        dir[0] = '.';
        dir[1] = '\0';
    }
    if (rf_str_contain_wild_cards(dir)) {
        rf_set_errno(RF_EINVAL);
        return NULL;
    }

    fd = rf_opendir_i(dir);
    if (fd == -1)
        return NULL;

    fctx = malloc(sizeof(rf_find_ctx_t));
    if (fctx == NULL) {
        rf_closedir_i(fd);
        rf_set_errno(RF_ENOMEM);
        return NULL;
    }
    fctx->fd = fd;
    fctx->pattern = strdup(p);
    if (ptr == NULL)
        return fctx;

    ret = rf_findnext(fctx, ptr);
    if (ret < 0) {
        rf_findclose(fctx);
        fctx = NULL;
    }
    return fctx;
}

rf_int_t rf_findnext(rf_handle_t fh, struct rf_dirent* ptr)
{
    rf_find_ctx_t* fctx = (rf_find_ctx_ref_t)fh;
    rf_char_t buf[RF_MAX_FNAME];
    rf_dirent_t dir;
    rf_int_t ret;

    for (;;) {
        dir.name = buf;
        dir.len = sizeof(buf);
        ret = rf_readdir_ex(fctx->fd, fctx->pattern, &dir);
        if (ret)
            return ret;
        ptr->attrib = dir.attrib;
        ret = rf_iconv(buf, dir.type, ptr->d_name, sizeof(ptr->d_name), g_path_code_page);
        RF_PERROR("rf_iconv error\n", ret != 0, { return -1; });
        /* ʱ fatļ ֧ģʽƥ!!! */
        if (rf_str_in_pattern(ptr->d_name, fctx->pattern))
            break;
    }
    return 0;
}

rf_int_t rf_findclose(rf_handle_t fh)
{
    rf_find_ctx_t* fctx = (rf_find_ctx_ref_t)fh;
    rf_int_t ret;

    ret = rf_closedir_i(fctx->fd);
    rf_free((rf_char_t*)fctx->pattern);
    rf_free(fctx);
    return ret;
}

RF_DIR* rf_opendir(const rf_char_t* name)
{
    RF_DIR* dir = 0;

    if (name && name[0]) {
        rf_size_t base_length = strlen(name);
        const char* all = strchr("/\\", name[base_length - 1]) ? "*" : "/*";

        if ((dir = (RF_DIR*)malloc(sizeof * dir)) != 0 &&
                (dir->name = (char*)malloc(base_length + strlen(all) + 1)) != 0) {
            strcat(strcpy(dir->name, name), all);
            if ((dir->fd = rf_findfirst(dir->name, &dir->result)) != NULL) {
                dir->result.d_name[0] = '\0';
            } else {
                free(dir->name);
                free(dir);
                dir = 0;
            }
        } else {
            free(dir);
            dir = 0;
            rf_set_errno(RF_ENOMEM);
        }
    } else {
        rf_set_errno(RF_EINVAL);
    }
    return dir;
}

rf_int_t rf_closedir(RF_DIR* dir)
{
    rf_int_t result = -1;

    if (dir) {
        if (dir->fd != NULL)
            result = rf_findclose(dir->fd);
        free(dir->name);
        free(dir);
    }

    if (result == -1)
        rf_set_errno(RF_EBADF);
    return result;
}

struct rf_dirent* rf_readdir(RF_DIR* dir)
{
    struct rf_dirent* result = NULL;
    if (rf_findnext(dir->fd, &dir->result) == 0)
        result = &dir->result;
    return result;
}

rf_void_t rf_rewinddir(RF_DIR* dir)
{
    if (dir && dir->fd != NULL) {
        rf_findclose(dir->fd);
        dir->fd = rf_findfirst(dir->name, NULL);
        dir->result.d_name[0] = '\0';
    } else {
        rf_set_errno(RF_EBADF);
    }
}
