/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "../common/rf_commonx.h"
#include "rf_fs.h"
#include "rf_fd.h"
#include "rf_file.h"
#if RF_OPT_USE_FAT
#include "../fat/fat_api.h"
#endif
#if RF_OPT_USE_EXT2
#include "../ext2/ext2_api.h"
#endif
#if RF_OPT_USE_NTFS
#include "../ntfs/ntfs_api.h"
#endif
#if RF_OPT_FS_SAVE_DEV_NAME
#include <string.h>
#endif

#ifndef RF_ARCH_VERSION_STRING
#define RF_ARCH_VERSION_STRING  __rf_mstrcat6__(RF_ARCH_STRING, RF_ARCH_STRING_1, RF_ARCH_STRING_2, RF_ARCH_STRING_3, RF_ARCH_STRING_4, RF_ARCH_STRING_5)
#endif

#if RF_OPT_USE_MALLOC == 0 && RF_OPT_USE_NTFS && RF_OPT_MAX_SECTOR_SIZE < 1024
#error "RF_OPT_MAX_SECTOR_SIZE: ntfs minimum support sector is 1024"
#endif

#if RF_OPT_USE_RF_PART == 1 && (RF_OPT_USE_MBR == 0 && RF_OPT_USE_GPT == 0)
#error "error At least need to support a partition type"
#endif

#if RF_OPT_USE_GPT == 1 && (RF_OPT_GT_4G_SECTORS == 0)
#error "error GPT partition need RF_OPT_GT_4G_SECTORS == 1"
#endif

/* ļϵͳ豸 */
static rf_vol_ctx_t g_vol_ctx_table[RF_OPT_MAX_VOLUME];
/* ǰڷ */
static rf_char_t g_work_vol = '\0';
/* 汾 ڶ */
static rf_char_t const* vtag = RF_VERSION_SHORT_STRING;
#ifdef RF_DEBUG
static rf_bool_t g_fs_init_flag = FALSE;
#endif

#if RF_OPT_PATH_LOCAL == 1
#ifdef _RF_MAX_NAME
rf_char_t xdata g_open_name_buf[ _RF_MAX_NAME ];
rf_char_t xdata g_find_name_buf[ _RF_MAX_NAME ];
#endif
#endif

#if RF_OPT_USE_GETCWD
rf_char_t g_curdir[RF_MAX_PATH];
#endif

#if RF_OPT_USE_VERSION
rf_version_t* rf_version(rf_void_t)
{
    static rf_version_t version = { 0 };
    if (!version.build) {
        version.major = RF_OPT_VERSION_MAJOR;
        version.minor = RF_OPT_VERSION_MINOR;
        version.alter = RF_OPT_VERSION_ALTER;
        version.build = RF_OPT_VERSION_BUILD;
    }
    return &version;
}
#endif

rf_vol_ctx_t* find_fs_ctx(const rf_char_t* path)
{
    rf_int_t idx, i;
    rf_char_t root;
    rf_vol_ctx_t* fs;

    RF_ASSERTE("error: invalid path\n", path != NULL);
    root = path_get_root(path);
    /* · Ĭϵǰ̷ */
    if (root == 0)
        root = g_work_vol;
    /* ǰ̷Ч ѵǰ̷л һ̷ */
    if (root == 0) {
        for (i = 0; i < RF_OPT_MAX_VOLUME; i++) {
            fs = &g_vol_ctx_table[i];
            if (fs->open_flag) {
                g_work_vol = 'a' + i;
                root = g_work_vol;
            }
        }
    }
    if (root == 0)
        return NULL;

    idx = (rf_int_t)(root - 'a');
    RF_PERROR("error: invalid vol idx\n", idx < 0 || idx > RF_OPT_MAX_VOLUME, return NULL);
    return &g_vol_ctx_table[idx];
}

#if RF_OPT_USE_MKFS
rf_int_t rf_imp_mkfs(const rf_char_t* fs_type, const rf_char_t* device_name, const rf_void_t* ctx)
{
    RF_PERROR("error: invalid param\n", !fs_type || !device_name, return RF_EINVAL);
    do {
#if RF_OPT_USE_FAT
        if (!strcmp(fs_type, "fat"))
            return fat_mkfs(device_name, ctx);
#endif
    } while (FALSE);
    return RF_EPERM;
}
#endif

rf_int_t rf_imp_mount(const rf_char_t* path, const rf_char_t* fs_type, rf_mount_param_t* param)
{
    rf_vol_ctx_t* fs = NULL;
    rf_int_t ret = RF_EINVAL;

    RF_PERROR("error: invalid param\n", (!param || !fs_type || !param->device_name || !param->ctx), return RF_EINVAL);
    fs = find_fs_ctx(path);
    RF_PERROR("error: invalid fs ctx\n", fs == NULL, return RF_EINVAL);
    RF_PERROR("error: Can't repeat mount\n", fs->open_flag == TRUE, return RF_EACCES);

    do {
#if RF_OPT_USE_FAT
        if (!strcmp(fs_type, "fat")) {
            ret = fat_mount(fs, param);
            break;
        }
#endif

#if RF_OPT_USE_NTFS
        if (!strcmp(fs_type, "ntfs")) {
            ret = ntfs_mount(fs, param);
            break;
        }
#endif

#if RF_OPT_USE_EXT2
        if (!strcmp(fs_type, "ext")) {
            ret = ext2_mount(fs, param);
            break;
        }
#endif
    } while (FALSE);
    if (ret != 0)
        return ret;

    /* һι õǰΪǰ */
    if (g_work_vol == '\0')
        g_work_vol = path[0];
#if RF_OPT_FS_SAVE_DEV_NAME
    fs->device_name = strdup(param->device_name);
#endif
    fs->open_flag = TRUE;
    return 0;
}

static rf_int_t rf_imp_unmount_i(rf_vol_ctx_t* fs)
{
    rf_int_t ret;

    FS_CHECK(fs);
    switch (((rf_vol_comm_t*)fs->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        ret = fat_unmount(fs);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        ret = ntfs_unmount(fs);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        ret = ext2_unmount(fs);
        break;
#endif

    default:
        return RF_ERROR;
    }
    if (ret != 0)
        return ret;
#if RF_OPT_FS_SAVE_DEV_NAME
    rf_free((rf_void_t*)fs->device_name);
#endif

    memset(fs, 0, sizeof(rf_vol_ctx_t));
    return 0;
}

rf_int_t rf_imp_unmount(const rf_char_t* path)
{
    rf_vol_ctx_t* fs = NULL;
    rf_int_t ret;

    fs = find_fs_ctx(path);
    ret = rf_imp_unmount_i(fs);
    if (ret)
        return ret;

    /* жصķΪ ǰ õǰΪЧ ҪĽ!!! */
    if (g_work_vol == path[0])
        g_work_vol = '\0';
    return 0;
}

static rf_int_t rf_imp_ioctl_i(rf_vol_ctx_t* fs, rf_int_t cmd, rf_void_t* buf)
{
    FS_CHECK(fs);

    switch (((rf_vol_comm_t*)fs->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_ioctl_vol(fs, cmd, buf);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_ioctl_vol(fs, cmd, buf);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_ioctl_vol(fs, cmd, buf);
        break;
#endif

    default:
        return RF_ERROR;
    }
}

#if RF_OPT_USE_STATFS
rf_int_t rf_imp_statfs(const rf_char_t* path, struct rf_statfs* buf)
{
    rf_vol_ctx_t* fs;

    RF_PERROR("error: invalid param\n", buf == NULL, return RF_EINVAL);
    fs = find_fs_ctx(path);
    return rf_imp_ioctl_i(fs, RF_VOL_IOCTL_STATFS, buf);
}
#endif

#if RF_OPT_USE_CHDIR
rf_int_t rf_imp_chdir(const rf_char_t* path)
{
    rf_vol_ctx_t* fs;
    rf_int_t ret;
    rf_char_t root;

    fs = find_fs_ctx(path);
    /* ·תΪ · д ٽл */
    ret = rf_imp_ioctl_i(fs, RF_VOL_IOCTL_CHDIR, (rf_void_ref_t)path);
    if (ret)
        return ret;

#if RF_OPT_USE_GETCWD
    /* ת·Ϊ · */
    ret = path_full(g_curdir, RF_MAX_PATH, path);
    if (ret) {
        /* תʧ ѵǰĿ¼ лĿ¼ */
        g_curdir[0] = RF_PATH_DEF_SEP_CHR;
        g_curdir[1] = '\0';
        ret = rf_imp_ioctl_i(fs, RF_VOL_IOCTL_CHDIR, (rf_void_ref_t)g_curdir);
        /* ٴʧ ǰ̷ΪЧ̷ */
        if (ret) {
            g_work_vol = '\0';
            return ret;
        }
    }
#endif

    /* ı䵱ǰ̷ */
    root = path_get_root(path);
    if (root != 0)
        g_work_vol = root;
    return 0;
}
#endif

#if RF_OPT_USE_GETCWD
rf_int_t rf_imp_getcwd(rf_char_t* path, rf_size_t len)
{
    RF_PERROR("error: invalid param\n", path == NULL || len < 4, return RF_EINVAL);

    path[0] = g_work_vol;
    /* ǰЧ̷ ѡЧ̷*/
    if (g_work_vol == '\0')
        find_fs_ctx("");
    /* ٴλȡʧʱ ֱӷ */
    if (g_work_vol == '\0')
        return 0;
    path[1] = ':';
    path[2] = '\0';
    return rf_strcat_s(path, (rf_int_t)len, g_curdir);
}
#endif

static rf_int_t rf_imp_init(void)
{
    rf_int_t ret = 0;

#ifdef RF_DEBUG
    /* ظʼ */
    RF_ASSERTE("repeated call rf_init\n", g_fs_init_flag == FALSE);
#endif
    /* ʹСǷ */
    RF_ASSERT_STATIC(sizeof(rf_u8_t) == 1);
    RF_ASSERT_STATIC(sizeof(rf_u16_t) == 2);
    RF_ASSERT_STATIC(sizeof(rf_u32_t) == 4);
    RF_ASSERT_STATIC(sizeof(rf_u64_t) == 8);
    RF_ASSERT_STATIC(sizeof(rf_char_t) == 1);
    RF_ASSERT_STATIC(sizeof(rf_wchar_t) == 2);
    /* С Ƿ */
    RF_ASSERTE("RF_OPT_CPU_BIG_ENDIAN set error\n", check_order_word());

    /* Ĭϵǰڷ Ч */
    g_work_vol = '\0';
    memset(&g_vol_ctx_table, 0, sizeof(g_vol_ctx_table));

    /* Ϊ̬ʹ ʼ̴ */
#if RF_OPT_DYNAMIC_LIB == 0
    rf_disk_init();
#endif

#if RF_OPT_USE_GETCWD
    /* ĬΪĿ¼ */
    g_curdir[0] = RF_PATH_DEF_SEP_CHR;
    g_curdir[1] = '\0';
#endif

    /* ʼfd */
    fd_init();

#if RF_OPT_USE_FAT
    fat_init();
#endif

#if RF_OPT_USE_EXT2
    ext2_init();
#endif

#if RF_OPT_USE_NTFS
    ntfs_init();
#endif

#if RF_OPT_USE_STDIO
    ret = rf_stdio_init();
#endif

#ifdef RF_DEBUG
    g_fs_init_flag = TRUE;
#endif

    return ret;
}

rf_int_t rf_init(void)
{
    rf_int_t ret;
    ret = rf_imp_init();
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}

#if RF_OPT_USE_RFS_EXIT
rf_int_t rf_imp_exit(rf_void_t)
{
    rf_int_t i, ret;
    rf_file_ctx_t* ctx;
    rf_vol_ctx_t* fs;

#if RF_OPT_USE_STDIO
    rf_stdio_exit();
#endif

    /* رļ */
    for (i = 0; i < fd_get_max_fd(); i++) {
        ctx = fd_get(i);
        if (ctx && ctx->open_flag) {
            ret = rf_file_close(ctx);
            RF_ASSERTE("error: close fd\n", ret != 0);
            fd_free(ctx);
        }
    }
    /* ͷfdʹõԴ */
    fd_exit();

    /* ж̷ */
    for (i = 0; i < RF_OPT_MAX_VOLUME; i++) {
        fs = &g_vol_ctx_table[i];
        if (fs->open_flag) {
            ret = rf_imp_unmount_i(fs);
            RF_ASSERTE("error: unmount\n", ret != 0);
            fs->open_flag = FALSE;
        }
    }

#ifdef RF_DEBUG
    g_fs_init_flag = FALSE;
#endif
    return 0;
}

rf_int_t rf_exit(void)
{
    rf_int_t ret;

    ret = rf_imp_exit();
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}
#endif
