/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_file.h"
#include "rf_stream.h"

rf_int_t _rf_nstream;
void** __rf_piob;

RF_FILE* rf_get_stream(void)
{
    RF_FILE* retval = NULL;
    rf_int_t i;

    /*
     * Loop through the __rf_piob table looking for a free stream, or the
     * first NULL entry.
     */
    for (i = 0 ; i < _rf_nstream ; i++) {

        if (__rf_piob[i] != NULL) {
            /*
             * if the stream is not inuse, return it.
             */
            if (!rf_inuse((RF_FILE*)__rf_piob[i])) {
                retval = (RF_FILE*)__rf_piob[i];
                break;
            }
        } else {
            /*
             * allocate a new _FILEX, set _piob[i] to it and return a
             * pointer to it.
             */
            if ((__rf_piob[i] = rf_malloc(sizeof(RF_FILE))) != NULL) {
                retval = (RF_FILE*)__rf_piob[i];
            }

            break;
        }
    }

    /*
     * Initialize the return stream.
     */
    if (retval != NULL) {
        retval->_flag = 0;
        retval->_cnt = 0;
        retval->_ptr = retval->_base = NULL;
        retval->_file = -1;
    }
    return (retval);
}

rf_void_t rf_free_stream(RF_FILE* stream)
{
    /*_ASSERTE(stream != NULL); */
    if (rf_inuse(stream) && rf_mbuf(stream)) {
        rf_free(stream->_base);
        stream->_flag &= ~(RF_IOMYBUF | RF_IOSETVBUF);
        stream->_base = stream->_ptr = NULL;
        stream->_cnt = 0;
    }
}

static rf_int_t rf_flush_stream(RF_FILE* str)
{
    RF_FILE* stream;
    rf_int_t rc = 0; /* assume good return */
    rf_int_t nchar;

    /* Init pointer to stream */
    stream = str;
    if ((stream->_flag & (RF_IOREAD | RF_IOWRT)) == RF_IOWRT && rf_bigbuf(stream)
            && (nchar = (rf_int_t)(stream->_ptr - stream->_base)) > 0) {
        if (rf_write(rf_fileno(stream), stream->_base, nchar) == nchar) {
            /* if this is a read/write file, clear _IOWRT so that
            * next operation can be a read
            */
            if (RF_IORW & stream->_flag)
                stream->_flag &= ~RF_IOWRT;
        } else {
            stream->_flag |= RF_IOERR;
            rc = EOF;
        }
    }

    stream->_ptr = stream->_base;
    stream->_cnt = 0;
    return (rc);
}

rf_void_t rf_getbuf(RF_FILE* str)
{
    RF_FILE* stream;

    /* Init pointers */
    stream = str;
    stream->_base = rf_malloc(RF_INTERNAL_BUFSIZ);
    if (stream->_base != NULL) {
        /* Got a big buffer */
        stream->_flag |= RF_IOMYBUF;
        stream->_bufsiz = RF_INTERNAL_BUFSIZ;
    } else {
        /* Did NOT get a buffer - use single char buffering. */
        stream->_flag |= RF_IONBF;
        stream->_base = (char*) & (stream->_charbuf);
        stream->_bufsiz = 1;
    }
    stream->_ptr = stream->_base;
    stream->_cnt = 0;
    return;
}

rf_int_t rf_setvbuf(RF_FILE* str, char* buffer, rf_int_t type, size_t size)
{
    RF_FILE* stream;
    rf_int_t retval = 0; /* assume good return */

    /*
     * (1) Make sure type is one of the three legal values.
     * (2) If we are buffering, make sure size is greater than 0.
     */
    if ((type != RF_IONBF) && ((size < 2) || (size > INT_MAX) ||
                               ((type != RF_IOFBF) && (type != RF_IOLBF))))
        return (-1);

#ifdef _WIN32
    /*
     * force size to be even by masking down to the nearest multiple
     * of 2
     */
    size &= (size_t)~1;
#endif  /* _WIN32 */

    /*
     * Init stream pointers
     */
    stream = str;

    /*
     * Flush the current buffer and free it, if it is ours.
     */
    rf_flush_stream(stream);
    rf_free_stream(stream);

    /*
     * Clear a bunch of bits in stream->_flag (all bits related to
     * buffering and those which used to be in stream2->_flag2). Most
     * of these should never be set when setvbuf() is called, but it
     * doesn't cost anything to be safe.
     */
    stream->_flag &= ~(RF_IOMYBUF | RF_IOYOURBUF | RF_IONBF |
#ifdef _WIN32
                       RF_IOSETVBUF |
#endif  /* _WIN32 */
                       RF_IOFEOF | RF_IOFLRTN | RF_IOCTRLZ);

    /*
     * CASE 1: No Buffering.
     */
    if (type & RF_IONBF) {
        stream->_flag |= RF_IONBF;
        buffer = (char*) & (stream->_charbuf);
        size = 1;
    }

    /*
     * NOTE: Cases 2 and 3 (below) cover type == _IOFBF or type == _IOLBF
     * Line buffering is treated as the same as full buffering, so the
     * _IOLBF bit in stream->_flag is never set. Finally, since _IOFBF is
     * defined to be 0, full buffering is simply assumed whenever _IONBF
     * is not set.
     */

    /*
     * CASE 2: Default Buffering -- Allocate a buffer for the user.
     */
    else if (buffer == NULL) {
        if ((buffer = rf_malloc(size)) == NULL) {
            retval = -1;
            goto done;
        }
        stream->_flag |= RF_IOMYBUF | RF_IOSETVBUF;
        stream->_flag |= RF_IOMYBUF;
    }

    /*
     * CASE 3: User Buffering -- Use the buffer supplied by the user.
     */
    else {
        stream->_flag |= RF_IOYOURBUF | RF_IOSETVBUF;
    }

    /*
     * Common return for all cases.
     */
    stream->_bufsiz = size;
    stream->_ptr = stream->_base = buffer;
    stream->_cnt = 0;
done:
    return (retval);
}

rf_int_t rf_filbuf(RF_FILE* str)
{
    RF_FILE* stream;

    /* Init pointer to _iob2 entry. */
    stream = str;
    if (!rf_inuse(stream) || stream->_flag & RF_IOSTRG)
        return -1;

    if (stream->_flag & RF_IOWRT) {
        stream->_flag |= RF_IOERR;
        return -1;
    }

    stream->_flag |= RF_IOREAD;

    /* Get a buffer, if necessary. */
    if (!rf_anybuf(stream))
        rf_getbuf(stream);
    else
        stream->_ptr = stream->_base;

    stream->_cnt = rf_read(rf_fileno(stream), stream->_base, stream->_bufsiz);
    if ((stream->_cnt == 0) || (stream->_cnt == -1)) {
        stream->_flag |= stream->_cnt ? RF_IOERR : RF_IOEOF;
        stream->_cnt = 0;
        return -1;
    }

    /*
    if (  !(stream->_flag & (RF_IOWRT|RF_IORW)) &&
          ((_osfile_safe(_fileno(stream)) & (FTEXT|FEOFLAG)) ==
            (FTEXT|FEOFLAG)) )
            stream->_flag |= RF_IOCTRLZ;*/

    /* Check for small _bufsiz (_SMALL_BUFSIZ). If it is small and
       if it is our buffer, then this must be the first _filbuf after
       an fseek on a read-access-only stream. Restore _bufsiz to its
       larger value (_INTERNAL_BUFSIZ) so that the next _filbuf call,
       if one is made, will fill the whole buffer. */
    if ((stream->_bufsiz == RF_SMALL_BUFSIZ) && (stream->_flag&
            RF_IOMYBUF) && !(stream->_flag & RF_IOSETVBUF)) {
        stream->_bufsiz = RF_INTERNAL_BUFSIZ;
    }
    stream->_cnt--;
    return (0xff & *stream->_ptr++);
}
